#!/usr/bin/env python3
# -*- coding: utf-8; -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import tbox
import math

import numpy as np
import os
import shutil

import params.src.write as pwrite
import params.src.indices as indices

import fwk

verb = False  # set to True for (a lot of) debug info

try:
    import mpi4py.MPI as mpi
    comm = mpi.COMM_WORLD
    rank = comm.rank
    siz  = comm.size
    name = mpi.Get_processor_name()
    status = mpi.Status()
    print("info: MPI found")
except:
    comm = None
    rank = 0
    siz  = 1
    name = "noname"
    print("info: MPI not found => MPI disabled")

def enum(*sequential, **named):
    enums = dict(list(zip(sequential, list(range(len(sequential))))), **named)
    return type('Enum', (), enums)

tags = enum('READY', 'DONE', 'EXIT', 'START', 'WAKEUP', 'SUICIDE')



class Job:
    """ Class containing data transmitted between MPI procs
    """
    def __init__(self, parameters, name, xi, m_i):
        # input
        self.m_i = m_i
        self.name = name
        self.parameters = parameters
        self.xi = xi
        # output
        self.QoI = 0.

    def execute(self):
        """
        [executed by worker processes]
        solve a given job and calculate "QoI"
        """
        import imp
        model = imp.load_source('Metafor_call',self.parameters['Metafor_model']+'/model.py')

        name = self.name
        if os.path.exists(name):
            shutil.rmtree(name)
        os.mkdir(name)
        os.chdir(name)

        self.QoI = model.model(self.parameters)

        os.chdir('..')


class Master:
    """
    MPI Process with rank #0
    """
    def __init__(self):
        self.slaves = list(range(1,siz))

    def start(self,parameters, surrogate_parameters, images_parameters):
        """
        master loop
        """
        global rank
        if rank!=0:
            raise Exception("this routine should be called with MPI rank=0")
        if verb: print("[%d] starting master" % rank)

        ##-----------------------------------------

        for s in self.slaves:
            if verb: print("[%d] sending wake-up signal to worker %d" % (rank,s))
            comm.send(None, dest=s, tag=tags.WAKEUP)

        #################################################
        ## Read the parameters
        fname = surrogate_parameters['fname']
        d = surrogate_parameters['d']
        M = surrogate_parameters['M']
        N = surrogate_parameters['N']

        Dom_min = surrogate_parameters['Dom_min']
        Dom_max = surrogate_parameters['Dom_max']
        P_name = surrogate_parameters['P_name']
        solver = surrogate_parameters['solver']

        normalisation = surrogate_parameters['normalisation']
        ##
        if surrogate_parameters['function_type'] == 'monomial':
            from params.src.monomials import phi
            from params.src.monomials import compute_n_max
            n_max = compute_n_max(N,d)
        elif surrogate_parameters['function_type'] == 'monomial_1norm':
            from params.src.monomials_1norm import phi
            from params.src.monomials_1norm import compute_n_max
            n_max = compute_n_max(N,d)
            print(n_max)
        if surrogate_parameters['training_points'] == 'gauss':
            import params.src.gauss as tp
        elif surrogate_parameters['training_points'] == 'gamma':
            import params.arnst_ponthot.src.gamma as tp
        ##

        m_max = 1
        for j in range(0,d):
            m_max = m_max*M[j]

        if surrogate_parameters['training_points'] == 'precomputed':


            import codecs, json

            #obj_text = codecs.open(surrogate_parameters['precomputed_training_points'], 'r', encoding='utf-8').read()
            #b_new = json.loads(obj_text)
            #X = np.array(b_new)

            X = np.genfromtxt(surrogate_parameters['precomputed_training_points'], dtype=None)

            #X = np.loadtxt('/home/kliegeois/Codes/waves/params/CVT/python/cvt_circle7.txt')
            tmp = np.shape(X)
            m_max = tmp[0]
            w = np.ones((m_max,))
        else:
            X, w = tp.grid(M)

        print(X)



        i =      np.zeros((d,), dtype =np.int)
        i_zero = np.zeros((d,), dtype =np.int)
        i_one =  np.ones((d,), dtype =np.int)
        p = 0

        xi = np.zeros((d,))

        if surrogate_parameters['display']:
            import params.src.plot as plot
            fig, ax, current_title = plot.init2(d, images_parameters)

            if not os.path.exists('images'):
                os.mkdir('images')
            else:
                shutil.rmtree('images')
                os.mkdir('images')

        m_i = 0
        sol = np.zeros((m_max,1))

        Z = np.zeros((m_max,n_max))
        W_sr = np.zeros((m_max,m_max))

        #################################################

        num_workers = siz-1
        closed_workers = 0
        while closed_workers<num_workers:
            # get a msg from any source
            data = comm.recv(source=mpi.ANY_SOURCE,
                             tag=mpi.ANY_TAG, status=status)
            source = status.Get_source()
            tag = status.Get_tag()
            if tag == tags.READY:
                if verb: print("[0] worker %d is ready" % source)

                #################################################

                if np.all(i != -i_one):
                    if surrogate_parameters['training_points'] == 'precomputed':
                        if normalisation == False:
                            for j in range(0,d):
                                print(X[0][0])
                                xi[j] = X[i[j]][j]
                        else:
                            means = surrogate_parameters['means']
                            variances = surrogate_parameters['variances']
                            correlations = surrogate_parameters['correlations']

                            for j in range(0,d):
                                xi[j] = (X[m_i,j]-means[j]) / (surrogate_parameters['nstddeviation']*np.sqrt(variances[j]))

                    elif surrogate_parameters['training_points'] == 'gamma':
                        means = surrogate_parameters['means']
                        variances = surrogate_parameters['variances']
                        correlations = surrogate_parameters['correlations']

                        xi_in = np.zeros((d,1))
                        for j in range(0,d):
                            xi_in[j] = X[i[j],j]
                        xi = tp.f_mapping(xi_in,d,means,variances,correlations,surrogate_parameters['f_mapping_method'])


                        print(xi)
                    elif normalisation == True:
                        for j in range(0,d):
                            xi[j] = X[i[j],j]
                    elif normalisation == False:
                        for j in range(0,d):
                            xi[j] = X[i[j],j]
                            xi[j] = Dom_min[j] + (Dom_max[j] - Dom_min[j])*(xi[j]+1.)/2.
                            parameters[P_name[j]] = xi[j]

                    # -------------------

                    if surrogate_parameters['training_points'] == 'precomputed':
                        Z[m_i,:] = np.transpose(phi(xi,N,n_max,d))
                        #print xi2
                    elif surrogate_parameters['training_points'] == 'gamma':
                        xi2 = np.zeros((d,))
                        for j in range(0,d):
                            xi2[j] = (xi[j][0]-means[j]) / np.sqrt(variances[j])
                        Z[m_i,:] = np.transpose(phi(xi2,N,n_max,d))
                        #print xi2
                    else:
                        Z[m_i,:] = np.transpose(phi(xi,N,n_max,d))


                    tmp = 1
                    if surrogate_parameters['training_points'] != 'precomputed':
                        for j in range(0,d):
                            tmp = tmp*w[i[j],j]

                    W_sr[m_i,m_i]  = np.sqrt(tmp)

                    # -------------------

                    if surrogate_parameters['training_points'] != 'precomputed':
                        for j in range(0,d):
                            tmp = xi[j]
                            print(tmp)
                            parameters[P_name[j]] = tmp

                    if surrogate_parameters['training_points'] == 'precomputed':
                        for j in range(0,d):
                            parameters[P_name[j]] =  X[m_i][j]

                    elif surrogate_parameters['training_points'] == 'gamma':
                        for j in range(0,d):
                            tmp = xi[j]
                            parameters[P_name[j]] = tmp[0]

                        #print parameters
                    elif normalisation == True:
                        for j in range(0,d):
                            xi[j] = Dom_min[j] + (Dom_max[j] - Dom_min[j])*(xi[j]+1.)/2.
                            parameters[P_name[j]] = xi[j]

                    # -------------------

                    name = 'i_'
                    for j in range(0,d-1):
                        name = name + str(i[j]) + '_'
                    name = name + str(i[d-1]) + '_'
                    name = name + 'M_'
                    for j in range(0,d-1):
                        name = name + str(M[j]) + '_'
                    name = name + str(M[d-1])

                    # -------------------

                    job = Job(parameters, name, xi, m_i)

                    if verb: print("[0] sending job to %d" % source)
                    comm.send(job, dest=source, tag=tags.START)

                    # -------------------

                    p,i = indices.my_next(p,i,M,d)
                    m_i = m_i+1
                else:

                    #################################################

                    if verb: print("[%d] exit %d" % (rank,source))
                    comm.send(None, dest=source, tag=tags.EXIT)

            elif tag==tags.DONE:
                if verb: print("[0] worker %d gives me its results" % source)

                #################################################
                m_ii = data.m_i
                xi = data.xi
                sol[m_ii] = data.QoI

                if surrogate_parameters['display']:
                    plot.scatter(xi,sol[m_ii],d,fig,ax, True,m_ii)

                #################################################

            elif tag==tags.EXIT:
                closed_workers+=1
                if verb: print("[0] worker %d exited (%d worker(s) still running)" % (source, num_workers-closed_workers))

        ##-----------------------------------------

        self.killslaves()


        Image = {}
        if surrogate_parameters['display']:
            Image['fig'] = fig
            Image['ax'] = ax
            Image['current_title'] = current_title
        return sol, Z, W_sr, Image

    def killslaves(self):
        global rank
        for s in self.slaves:
            if verb: print("[%d] sending suicide job to %d" % (rank,s))
            comm.send(None, dest=s, tag=tags.SUICIDE)


class Worker:
    """
    MPI Process with rank #1-#n
    """
    def __init__(self, rank):
        self.rank = rank

    def start(self):
        """
        worker loop
        """
        if self.rank==0:
            raise Exception("this routine should be called with MPI rank!=0")

        if verb: print("[%d] starting worker" % self.rank)
        timers = fwk.Timers()
        timers['total'].start()
        while True:
            comm.recv(source=0, tag=mpi.ANY_TAG, status=status)
            tag = status.Get_tag()
            if tag==tags.WAKEUP:
                if verb: print("[%d] waking up" % self.rank)
                while True:
                    comm.send(None, dest=0, tag=tags.READY)
                    job = comm.recv(source=0, tag=mpi.ANY_TAG, status=status)
                    tag = status.Get_tag()
                    if tag==tags.START:
                        if verb: print("[%d] starting job" % self.rank)
                        name = str(job.m_i)+'_'+job.name
                        timers[name].start()
                        job.execute()
                        timers[name].stop()
                        if verb: print("[%d] sending job results" % self.rank)
                        comm.send(job, dest=0, tag=tags.DONE)
                    elif tag==tags.EXIT:
                        if verb: print("[%d] sending exit confirmation" % self.rank)
                        comm.send(None, dest=0, tag=tags.EXIT)
                        timers['total'].stop()
                        print(timers)
                        break
            elif tag==tags.SUICIDE:
                if verb: print("[%d] I'm dying..." % self.rank)
                break
