#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute flow around the Onera M6 wing at 3 degrees AOA and Mach 0.84 for benchmark
# Adrien Crovato

import numpy as np
import flow.default as floD
import flow.utils as floU
import tbox.utils as tbxU
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

try:
   import tboxVtk
   BestWriter = tboxVtk.VtkExport
   canPost = True
except:
   BestWriter = tbox.GmshExport
   canPost = False

def newton(pbl):
    from fwk.wutils import parseargs
    import flow
    import tbox
    # Pardiso and GMRES should give similar perfs
    try:
        newton = flow.Newton(pbl, tbox.Pardiso())
    except:
        gmres = tbox.Gmres()
        gmres.setFillFactor(2)
        gmres.setDropTol(1e-6)
        gmres.setRestart(50)
        gmres.setTolerance(1e-5)
        newton = flow.Newton(pbl, gmres)
    newton.nthreads = parseargs().k
    newton.verbose = 1
    newton.relTol = 1e-6
    newton.absTol = 1e-8
    newton.lsTol = 1e-6
    newton.avThrsh = 1e-2
    return newton

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 3.06*np.pi/180
    M_inf = 0.839
    M_crit = 5 # Squared critical Mach number (above which density is modified)
    dim = 3

    # define dimension
    S_ref = 0.7528 # reference area
    c_ref = 0.64 # reference chord (MAC)

    # mesh the wing
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    msh, writer = floD.mesh(dim, 'benchmark/onera_lfs.msh', {}, ['field', 'wing', 'symmetry', 'downstream'], wktp = 'wakeTip')
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, _, _ = floD.problem(msh, dim, alpha, 0., M_inf, M_crit, S_ref, c_ref, 0., 0., 0., 'wing', tp = 'teTip')
    tms['pre'].stop()

    # solve problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    solver = newton(pbl)
    solver.run()
    solver.save(writer)
    tms['solver'].stop()

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('CL', solver.Cl, 0.29, 1e-1))
    tests.add(CTest('CD', solver.Cd, 0.011, 1e-1))
    tests.add(CTest('CS', solver.Cs, 0.014, 1e-1))
    tests.add(CTest('CM', solver.Cm, -0.212, 1e-1))
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
