/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wLoadFunctional.h"
#include "wFct0.h"

#include "wElement.h"
#include "wCache.h"
#include "wGauss.h"
#include "wMem.h"

using namespace tbox;
using namespace flow;

/**
 * @brief Compute the loads due to pressure normalized by the freestream dynamic pressure, on one boundary element
 *
 * b = \int psi * cp * n dS
 */
Eigen::VectorXd LoadFunctional::build(Element const &e, Element const &eV, std::vector<double> const &phi, Fct0 const &cp)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();
    Mem &mem = e.getVMem();

    // Pressure force coefficient in associated volume
    Eigen::Vector3d fcp = cp.eval(eV, phi, 0) * e.normal();

    // Integrate cp on the surface
    Eigen::VectorXd b = Eigen::VectorXd::Zero(3 * e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        Eigen::VectorXd sf = cache.getSf(k);
        double wdj = gauss.getW(k) * mem.getDetJ(k);
        for (size_t i = 0; i < e.nodes.size(); ++i)
            b.block<3, 1>(i * 3, 0) += sf(i) * fcp * wdj; // stack nodal force vectors [fx1, fy1, fz1, fx2, fy2, fz2, ...]
    }
    return b;
}

/**
 * @brief Build the gradient of the loads with respect to the flow variable, on one boundary element
 *
 * A = \int psi * dcp * n dS
 */
Eigen::MatrixXd LoadFunctional::buildGradientFlow(Element const &e, Element const &eV, std::vector<double> const &phi, Fct0 const &cp)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();
    Mem &mem = e.getVMem();

    // Gradient of pressure force coefficient in associated volume
    Eigen::MatrixXd dfcp = e.normal() * cp.evalD(eV, phi, 0) * eV.computeGrad(phi, 0).transpose() * eV.getVMem().getJinv(0) * eV.getVCache().getDsf(0);

    // Integrate cp on the surface
    Eigen::MatrixXd A = Eigen::MatrixXd::Zero(3 * e.nodes.size(), eV.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        Eigen::VectorXd sf = cache.getSf(k);
        double wdj = gauss.getW(k) * mem.getDetJ(k);
        for (size_t i = 0; i < e.nodes.size(); ++i)
            A.block(i * 3, 0, 3, eV.nodes.size()) += sf(i) * dfcp * wdj;
    }
    return A;
}

/**
 * @brief Build the gradient of the loads with respect to the nodes, on one boundary element
 *
 * A = d( \int psi * cp * n dS )
 *   = \int psi * dcp * n * dS
 *   + \int psi * cp * dn * dS
 *   + \int psi * cp * n * ddS
 */
std::tuple<Eigen::MatrixXd, Eigen::MatrixXd> LoadFunctional::buildGradientMesh(Element const &e, Element const &eV, std::vector<double> const &phi, Fct0 const &cp, int nDim)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();
    Mem &mem = e.getVMem();

    // Variables
    double cP = cp.eval(eV, phi, 0);  // pressure coefficient
    Eigen::Vector3d nrm = e.normal(); // unit normal
    // Gradients
    double dCp = cp.evalD(eV, phi, 0);                                 // pressure coefficient
    Eigen::VectorXd dPhi = eV.computeGrad(phi, 0);                     // potential
    Eigen::MatrixXd const &iJ = eV.getVMem().getJinv(0);               // inverse Jacobian
    std::vector<Eigen::MatrixXd> const &dJ = eV.getVMem().getGradJ(0); // Jacobian
    std::vector<Eigen::Vector3d> dNrm = e.gradientNormal();            // unit normal

    // Build element contributions
    Eigen::MatrixXd As = Eigen::MatrixXd::Zero(3 * e.nodes.size(), e.nodes.size() * nDim);
    Eigen::MatrixXd Av = Eigen::MatrixXd::Zero(3 * e.nodes.size(), eV.nodes.size() * nDim);
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        Eigen::VectorXd sf = cache.getSf(k);
        double w = gauss.getW(k);
        double detJ = mem.getDetJ(k);
        std::vector<double> dDetJ = mem.getGradDetJ(k);
        for (size_t i = 0; i < e.nodes.size(); ++i)
        {
            for (size_t m = 0; m < nDim; ++m)
            {
                // surface
                for (size_t j = 0; j < e.nodes.size(); ++j)
                {
                    size_t idx = j * nDim + m;
                    As.block<3, 1>(i * 3, idx) += sf(i) * cP * w * (dNrm[idx] * detJ + nrm * dDetJ[idx]); // Cp * dn * S + Cp * n * dS
                }
                // volume
                for (size_t j = 0; j < eV.nodes.size(); ++j)
                {
                    size_t idx = j * nDim + m;
                    Av.block<3, 1>(i * 3, idx) += sf(i) * dCp * dPhi.transpose() * (-iJ * dJ[idx] * dPhi) * nrm * w * detJ; // dCp * n * S
                }
            }
        }
    }
    return std::make_tuple(As, Av);
}
