/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wProblem.h"
#include "wElement.h"
#include "wMem.h"
#include "wMedium.h"
#include "wBoundary.h"
#include "wAssign.h"
#include "wFreestream.h"
#include "wWake.h"
#include "wKutta.h"
#include "wBlowing.h"
using namespace tbox;
using namespace flow;

Problem::Problem(std::shared_ptr<MshData> _msh, int dim, double aoa, double aos,
                 double minf, double sref, double cref, double xref,
                 double yref, double zref) : msh(_msh), nDim(dim), alpha(aoa),
                                             beta(aos), M_inf(minf), S_ref(sref),
                                             c_ref(cref)
{
    x_ref(0) = xref;
    x_ref(1) = yref;
    x_ref(2) = zref;
}

/**
 * @brief Compute drag, sideforce and lift normalized directions
 * @todo use F1ElVi
 */
std::tuple<Eigen::Vector3d, Eigen::Vector3d, Eigen::Vector3d> Problem::computeDirections()
{
    Eigen::Vector3d dirD, dirS, dirL;
    if (nDim == 2)
    {
        dirD = Eigen::Vector3d(cos(alpha), sin(alpha), 0) / S_ref;
        dirS = Eigen::Vector3d(0, 0, 0);
        dirL = Eigen::Vector3d(-sin(alpha), cos(alpha), 0) / S_ref;
    }
    else
    {
        dirD = Eigen::Vector3d(cos(alpha) * cos(beta), sin(beta), sin(alpha) * cos(beta)) / S_ref;
        dirS = Eigen::Vector3d(-cos(alpha) * sin(beta), cos(beta), -sin(alpha) * sin(beta)) / S_ref;
        dirL = Eigen::Vector3d(-sin(alpha), 0, cos(alpha)) / S_ref;
    }
    return std::make_tuple(dirD, dirS, dirL);
}

/**
 * @brief Compute gradient of drag, sideforce and lift normalized directions
 * @todo use F1ElVi
 * @todo only wrt alpha
 */
std::tuple<Eigen::Vector3d, Eigen::Vector3d, Eigen::Vector3d> Problem::computeGradientDirections()
{
    Eigen::Vector3d dirD, dirS, dirL;
    if (nDim == 2)
    {
        dirD = Eigen::Vector3d(-sin(alpha), cos(alpha), 0) / S_ref;
        dirS = Eigen::Vector3d(0, 0, 0);
        dirL = Eigen::Vector3d(-cos(alpha), -sin(alpha), 0) / S_ref;
    }
    else
    {
        dirD = Eigen::Vector3d(-sin(alpha) * cos(beta), sin(beta), cos(alpha) * cos(beta)) / S_ref;
        dirS = Eigen::Vector3d(sin(alpha) * sin(beta), cos(beta), -cos(alpha) * sin(beta)) / S_ref;
        dirL = Eigen::Vector3d(-cos(alpha), 0, -sin(alpha)) / S_ref;
    }
    return std::make_tuple(dirD, dirS, dirL);
}

/**
 * @brief Compute freestream velocity
 * @todo use F1ElVi
 */
Eigen::VectorXd Problem::computeVi()
{
    if (nDim == 2)
        return Eigen::Vector2d(cos(alpha), sin(alpha));
    else
        return Eigen::Vector3d(cos(alpha) * cos(beta), sin(beta), sin(alpha) * cos(beta));
}

/**
 * @brief Compute freestream velocity
 * @todo use F1ElVi
 * @todo only wrt alpha
 */
Eigen::VectorXd Problem::computeGradientVi()
{
    if (nDim == 2)
        return Eigen::Vector2d(-sin(alpha), cos(alpha));
    else
        return Eigen::Vector3d(-sin(alpha) * cos(beta), 0, cos(alpha) * cos(beta));
}

/**
 * @brief Set the fluid medium
 * @authors Adrien Crovato
 */
void Problem::set(std::shared_ptr<Medium> m)
{
    medium = m;
}

/**
 * @brief Add a boundary surface
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Boundary> b)
{
    bnds.push_back(b);
}

/**
 * @brief Add initial condition
 * @authors Adrien Crovato
 */
void Problem::set(std::shared_ptr<Initial> i)
{
    iIC = i;
}

/**
 * @brief Add Dirichlet boundary condition
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Dirichlet> d)
{
    dBCs.push_back(d);
}

/**
 * @brief Add freestream (Neumann) boundary condition
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Freestream> f)
{
    fBCs.push_back(f);
}

/**
 * @brief Add wake boundary condition
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Wake> w)
{
    wBCs.push_back(w);
}

/**
 * @brief Add Kutta condition
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Kutta> k)
{
    kSCs.push_back(k);
}

/**
 * @brief Add Blowing boundary condition
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Blowing> b)
{
    bBCs.push_back(b);
}

/**
 * @brief Initialize the elements memory (values)
 * @authors Adrien Crovato
 */
void Problem::initMem()
{
    // Update volume
    for (auto e : medium->tag->elems)
        e->getVMem().update(true);
    // Update surface (boundaries)
    for (auto surf : bnds)
        for (auto e : surf->groups[0]->tag->elems)
            e->getVMem().update(false);
    // Update surface (Freestream B.C.)
    for (auto surf : fBCs)
        for (auto e : surf->tag->elems)
            e->getVMem().update(false);
    // Update surface (Wake B.C.)
    for (auto surf : wBCs)
    {
        for (auto e : surf->groups[0]->tag->elems)
            e->getVMem().update(false);
        for (auto e : surf->groups[1]->tag->elems)
            e->getVMem().update(false);
    }
    // Update surface (Blowing B.C.) if not already done
    for (auto surf : bBCs)
        for (auto e : surf->tag->elems)
            if (!(e->getVMem().isValid()))
                e->getVMem().update(false);
}

/**
 * @brief Initialize the elements memory (gradients)
 * @authors Adrien Crovato
 */
void Problem::initGradMem()
{
    // Update volume
    for (auto e : medium->tag->elems)
        e->getVMem().initGradients(true);
    // Update surface (boundaries)
    for (auto surf : bnds)
        for (auto e : surf->groups[0]->tag->elems)
            e->getVMem().initGradients(false);
    // Update surface (Wake B.C.)
    for (auto surf : wBCs)
    {
        for (auto e : surf->groups[0]->tag->elems)
            e->getVMem().initGradients(false);
        for (auto e : surf->groups[1]->tag->elems)
            e->getVMem().initGradients(false);
    }
    // Update surface (Blowing B.C.) if not already done
    for (auto surf : bBCs)
        for (auto e : surf->tag->elems)
            if (!(e->getVMem().isGradientValid()))
                e->getVMem().initGradients(false);
}

/**
 * @brief Pin a degree of freedom if no Dirichlet BC is explicitely provided
 */
void Problem::pin()
{
    if (dBCs.empty())
        this->add(std::make_shared<Dirichlet>(this->msh, this->fBCs.front()->tag->no, this->medium->phiInf, true));
}

/**
 * @brief Check that Problem is not empty and that elements are supported
 * @authors Adrien Crovato
 */
void Problem::check() const
{
    // Sanity checks
    if (medium == NULL)
        throw std::runtime_error("No Medium provided!\n");
    if (iIC == NULL)
        throw std::runtime_error("No Initial Condition provided!\n");
    if (fBCs.empty())
        throw std::runtime_error("No Freestream B.C. provided!\n");
    if (wBCs.empty())
        throw std::runtime_error("No Wake B.C. provided!\n");
    // Three-dimension problem
    if (nDim == 3)
    {
        // Volume element type
        for (auto e : medium->tag->elems)
        {
            if (e->type() != ELTYPE::TETRA4)
            {
                std::stringstream err;
                err << "3D solver is only implemented for volume elements of type "
                    << ELTYPE::TETRA4 << " (" << e->type() << " was given)!\n";
                throw std::runtime_error(err.str());
            }
        }
        // Surface element type (Freestream B.C.)
        for (auto surf : fBCs)
        {
            for (auto e : surf->tag->elems)
            {
                if (e->type() != ELTYPE::TRI3)
                {
                    std::stringstream err;
                    err << "3D solver is only implemented for surface elements of type "
                        << ELTYPE::TRI3 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
        }
        // Surface element type (Wake B.C.)
        for (auto surf : wBCs)
        {
            for (auto e : surf->groups[0]->tag->elems)
            {
                if (e->type() != ELTYPE::TRI3)
                {
                    std::stringstream err;
                    err << "3D solver is only implemented for surface elements of type "
                        << ELTYPE::TRI3 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
            for (auto e : surf->groups[1]->tag->elems)
            {
                if (e->type() != ELTYPE::TRI3)
                {
                    std::stringstream err;
                    err << "3D solver is only implemented for surface elements of type "
                        << ELTYPE::TRI3 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
        }
        // Blowing B.C.
        if (!bBCs.empty())
            throw std::runtime_error("Blowing boundary conditions are not supported for 3D problems!\n");
    }
    // Two-dimension problem
    else if (nDim == 2)
    {
        // Volume element type
        for (auto e : medium->tag->elems)
        {
            if (e->type() != ELTYPE::TRI3)
            {
                std::stringstream err;
                err << "2D solver is only implemented for volume elements of type "
                    << ELTYPE::TRI3 << " (" << e->type() << " was given)!\n";
                throw std::runtime_error(err.str());
            }
        }
        // Surface element type (Freestream B.C.)
        for (auto surf : fBCs)
        {
            for (auto e : surf->tag->elems)
            {
                if (e->type() != ELTYPE::LINE2)
                {
                    std::stringstream err;
                    err << "2D solver is only implemented for surface elements of type "
                        << ELTYPE::LINE2 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
        }
        // Surface element type (Wake B.C.)
        for (auto surf : wBCs)
        {
            for (auto e : surf->groups[0]->tag->elems)
            {
                if (e->type() != ELTYPE::LINE2)
                {
                    std::stringstream err;
                    err << "2D solver is only implemented for surface elements of type "
                        << ELTYPE::LINE2 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
            for (auto e : surf->groups[1]->tag->elems)
            {
                if (e->type() != ELTYPE::LINE2)
                {
                    std::stringstream err;
                    err << "2D solver is only implemented for surface elements of type "
                        << ELTYPE::LINE2 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
        }
    }
    else
    {
        std::stringstream err;
        err << "Solver is only implemented for 2D or 3D problems (dimension "
            << nDim << " was given)!\n";
        throw std::runtime_error(err.str());
    }
}

void Problem::write(std::ostream &out) const
{
    out << "flow::Problem parameters"
        << "\n\tDimension: " << nDim
        << "\n\tAngle of attack: " << alpha
        << "\n\tAngle of sideslip: " << beta
        << "\n\tFreestream Mach number: " << M_inf
        << "\n\tReference area: " << S_ref
        << "\n\tReference length: " << c_ref
        << "\n\tReference point: [" << x_ref(0) << ", " << x_ref(1) << ", " << x_ref(2) << ']'
        << std::endl;
    out << "with"
        << "\n\t" << *medium;
    for (auto b : bnds)
        out << "\n\t" << *b;
    out << "\n\t" << *iIC;
    for (auto d : dBCs)
        out << "\n\t" << *d;
    for (auto f : fBCs)
        out << "\n\t" << *f;
    for (auto w : wBCs)
        out << "\n\t" << *w;
    for (auto k : kSCs)
        out << "\n\t" << *k;
    for (auto b : bBCs)
        out << "\n\t" << *b;
}
