/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WPROBLEM_H
#define WPROBLEM_H

#include "flow.h"
#include "wObject.h"
#include <memory>
#include <iostream>
#include <vector>
#include <Eigen/Dense>

namespace flow
{

/**
 * @brief Manage the problem
 *
 * Contains mesh, medium, boundaries and reference value
 * @authors Adrien Crovato
 */
class FLOW_API Problem : public fwk::wSharedObject
{
public:
    std::shared_ptr<tbox::MshData> msh; ///< Mesh structure
#ifndef SWIG
    std::shared_ptr<Medium> medium;              ///< Fluid
    std::vector<std::shared_ptr<Boundary>> bnds; ///< Boundaries

    std::shared_ptr<Initial> iIC;                  ///< Initial condition
    std::vector<std::shared_ptr<Dirichlet>> dBCs;  ///< Dirichlet boundary conditions
    std::vector<std::shared_ptr<Freestream>> fBCs; ///< Freestream boundary conditions
    std::vector<std::shared_ptr<Wake>> wBCs;       ///< Wake boundary condition
    std::vector<std::shared_ptr<Kutta>> kSCs;      ///< Kutta condition
    std::vector<std::shared_ptr<Blowing>> bBCs;    ///< Blowing (transpiration) boundary condition
#endif
    // Reference values
    int nDim;              ///< Problem dimension
    double alpha;          ///< Angle of attack
    double beta;           ///< Angle of sideslip
    double M_inf;          ///< Freestream Mach number
    double S_ref;          ///< Reference surface
    double c_ref;          ///< Reference chord
    Eigen::Vector3d x_ref; ///< Reference center point (for moment computation)

    Problem(std::shared_ptr<tbox::MshData> _msh, int dim, double aoa, double aos,
            double minf, double sref, double cref, double xref, double yref,
            double zref);
    virtual ~Problem() { std::cout << "~Problem()\n"; }

    std::tuple<Eigen::Vector3d, Eigen::Vector3d, Eigen::Vector3d> computeDirections();
    std::tuple<Eigen::Vector3d, Eigen::Vector3d, Eigen::Vector3d> computeGradientDirections();
    Eigen::VectorXd computeVi();
    Eigen::VectorXd computeGradientVi();
    void set(std::shared_ptr<Medium> m);
    void add(std::shared_ptr<Boundary> b);
    void set(std::shared_ptr<Initial> i);
    void add(std::shared_ptr<Dirichlet> d);
    void add(std::shared_ptr<Freestream> f);
    void add(std::shared_ptr<Wake> w);
    void add(std::shared_ptr<Kutta> k);
    void add(std::shared_ptr<Blowing> b);

#ifndef SWIG
    void check() const;
    void pin();
    void initMem();
    void initGradMem();
    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace flow

#endif //WPROBLEM_H
