#ifndef KATOPTRON_CONTACTSLIST_H
#define KATOPTRON_CONTACTSLIST_H

#include "katoptron.h"

#include "Map.h"
#include "NodesList.h"
#include "ElementsList.h"

#include <Tpetra_Map.hpp>
#include <Tpetra_Vector.hpp>
#include <Teuchos_RCP.hpp>
#include <Kokkos_ViewFactory.hpp>

namespace katoptron
{

/**
 * @brief Class which is used to store the list of the contact interfaces.
 */
class ContactsList
{
private:
    bool at_least_one_sticking = false;

    typedef typename Map::local_ordinal_type local_ordinal_type;
    typedef typename Map::global_ordinal_type global_ordinal_type;

    Kokkos::View<global_ordinal_type **, Kokkos::LayoutRight> contacts;
    Kokkos::View<global_ordinal_type **, Kokkos::LayoutRight> contacts_dirichlet;

    Kokkos::View<local_ordinal_type *, Kokkos::LayoutRight> shape_function;

    Kokkos::View<local_ordinal_type *, Kokkos::LayoutRight> signorini;

    Kokkos::View<local_ordinal_type *, Kokkos::LayoutRight> initially_open;

    Kokkos::View<local_ordinal_type *, Kokkos::LayoutRight> no_update;
    Kokkos::View<local_ordinal_type *, Kokkos::LayoutRight> sticking;
    Kokkos::View<local_ordinal_type *, Kokkos::LayoutRight> mesh_tying;

    Kokkos::View<double *, Kokkos::LayoutRight> contacts_scale;

    Kokkos::View<local_ordinal_type **, Kokkos::LayoutRight> initially_closed_nodes;

    Kokkos::View<double **, Kokkos::LayoutRight> contacts_normal;

    local_ordinal_type contacts_number;
    local_ordinal_type contacts_slave_nodes_size;
    local_ordinal_type contacts_slave_elems_size;
    local_ordinal_type contacts_master_nodes_size;
    local_ordinal_type contacts_master_elems_size;
    local_ordinal_type contacts_slave_normal;
    local_ordinal_type contacts_master_normal;
    local_ordinal_type contacts_slave_nodes;
    local_ordinal_type contacts_slave_elems;
    local_ordinal_type contacts_master_nodes;
    local_ordinal_type contacts_master_elems;

public:
    ContactsList(Problem &pbl, Teuchos::RCP<Map> map, Teuchos::RCP<ElementsList> elementsList,
                 Teuchos::RCP<NodesList> nodesList);

    ~ContactsList()
    {
        Kokkos::resize(contacts, 0, 0);
        Kokkos::resize(contacts_normal, 0, 0);
        Kokkos::resize(shape_function, 0);
        Kokkos::resize(signorini, 0);
        Kokkos::resize(initially_open, 0);
        Kokkos::resize(no_update, 0);
        Kokkos::resize(sticking, 0);
        Kokkos::resize(mesh_tying, 0);
        Kokkos::resize(contacts_scale, 0);
        Kokkos::resize(contacts_dirichlet, 0, 0);
        Kokkos::resize(initially_closed_nodes, 0, 0);
    }

    /**
    * @brief Return the number of contact interfaces.
    * 
    */
    inline local_ordinal_type getContactNumber() { return contacts_number; }

    /**
    * @brief Return the number of slave nodes for a
    * given interface.
    * 
    * Argument:
    * <ul>
    *  <li> i: the interface ID.
    * </ul>
    */
    inline local_ordinal_type getSlaveNodesSize(local_ordinal_type i)
    {
        return contacts(i, contacts_slave_nodes_size);
    }

    /**
    * @brief Return the number of slave elements for a
    * given interface.
    * 
    * Argument:
    * <ul>
    *  <li> i: the interface ID.
    * </ul>
    */
    inline local_ordinal_type getSlaveElementsSize(local_ordinal_type i)
    {
        return contacts(i, contacts_slave_elems_size);
    }

    /**
    * @brief Return the number of master nodes for a
    * given interface.
    * 
    * Argument:
    * <ul>
    *  <li> i: the interface ID.
    * </ul>
    */
    inline local_ordinal_type getMasterNodesSize(local_ordinal_type i)
    {
        return contacts(i, contacts_master_nodes_size);
    }

    /**
    * @brief Return the number of master elements for a
    * given interface.
    * 
    * Argument:
    * <ul>
    *  <li> i: the interface ID.
    * </ul>
    */
    inline local_ordinal_type getMasterElementsSize(local_ordinal_type i)
    {
        return contacts(i, contacts_master_elems_size);
    }

    /**
    * @brief Return the global ID of the slave node n1 of 
    * a given interface.
    * 
    * Arguments:
    * <ul>
    *  <li> i: the interface ID,
    *  <li> n1: the ID of the node on the current interface.
    * </ul>
    */
    inline global_ordinal_type getSlaveNode(local_ordinal_type i, local_ordinal_type n1)
    {
        return contacts(i, contacts_slave_nodes + n1);
    }

    /**
    * @brief Return the local ID of the slave element e1 of 
    * a given interface.
    * 
    * Arguments:
    * <ul>
    *  <li> i: the interface ID,
    *  <li> e1: the ID of the element on the current interface.
    * </ul>
    */
    inline local_ordinal_type getSlaveElement(local_ordinal_type i, local_ordinal_type e1)
    {
        return contacts(i, contacts_slave_elems + e1);
    }

    /**
    * @brief Return the global ID of the master node n1 of 
    * a given interface.
    * 
    * Argument:
    * <ul>
    *  <li> i: the interface ID,
    *  <li> n1: the ID of the node on the current interface.
    * </ul>
    */
    inline global_ordinal_type getMasterNode(local_ordinal_type i, local_ordinal_type n1)
    {
        return contacts(i, contacts_master_nodes + n1);
    }

    /**
    * @brief Return the local ID of the master element e1 of 
    * a given interface.
    * 
    * Arguments:
    * <ul>
    *  <li> i: the interface ID,
    *  <li> e1: the ID of the element on the current interface.
    * </ul>
    */
    inline local_ordinal_type getMasterElement(local_ordinal_type i, local_ordinal_type e1)
    {
        return contacts(i, contacts_master_elems + e1);
    }

    /**
    * @brief Return the jth component of the normal vector to the slave
    * side of a given interface.
    * 
    * Arguments:
    * <ul>
    *  <li> i: the interface ID,
    *  <li> j: the index of the component of the normal vector.
    * </ul> 
    */
    inline double getSlaveNormal(local_ordinal_type i, local_ordinal_type j)
    {
        return contacts_normal(i, j);
    }

    /**
    * @brief Return the jth component of the normal vector to the master
    * side of a given interface.
    * 
    * Arguments:
    * <ul>
    *  <li> i: the interface ID,
    *  <li> j: the index of the component of the normal vector.
    * </ul>
    */
    inline double getMasterNormal(local_ordinal_type i, local_ordinal_type j)
    {
        return contacts_normal(i, j + 3);
    }

    /**
    * @brief Return true if the shape functions of the Lagrange 
    * multipliers are the standard shape function for a given interface.
    * 
    * Argument:
    * <ul>
    *  <li> i: the interface ID.
    * </ul>
    */
    inline bool isStandardShapeFunction(local_ordinal_type i)
    {
        return shape_function(i) == 1;
    }

    /**
    * @brief Return true if contact of the given interface is 
    * initially open.
    * 
    * Argument:
    * <ul>
    *  <li> i: the interface ID.
    * </ul>
    */
    inline bool isInitiallyOpen(local_ordinal_type i)
    {
        return initially_open(i) == 1;
    }

    /**
    * @brief Return true if contact of the given interface is 
    * initially close.
    * 
    * Argument:
    * <ul>
    *  <li> i: the interface ID.
    * </ul>
    */
    inline bool isInitiallyClose(local_ordinal_type i)
    {
        return initially_open(i) == 0;
    }

    /**
     * @brief Return true if the contact status of the given interface 
     * should be checked and potentially updated.
     * 
     * Argument:
     * <ul>
     *  <li> i: the interface ID.
     * </ul>
     */
    inline bool isUpdated(local_ordinal_type i)
    {
        return no_update(i) == 0;
    }

    /**
     * @brief Return true if the contact status of the given interface 
     * should not be checked.
     * 
     * Argument:
     * <ul>
     *  <li> i: the interface ID.
     * </ul>
     */
    inline bool isNotUpdated(local_ordinal_type i)
    {
        return no_update(i) == 1;
    }

    /**
     * @brief Return true if the given interface is a sticking contact
     * interface.
     * 
     * Argument:
     * <ul>
     *  <li> i: the interface ID.
     * </ul>
     */
    inline bool isSticking(local_ordinal_type i)
    {
        return sticking(i);
    }

    /**
     * @brief Return true if the given interface is a mesh tying
     * interface.
     * 
     * Argument:
     * <ul>
     *  <li> i: the interface ID.
     * </ul>
     */
    inline bool isTying(local_ordinal_type i)
    {
        return mesh_tying(i);
    }

    /**
     * @brief Return true if at least one contact interface is sticking.
     */
    inline bool hasAtLeastOneSticking(void)
    {
        return at_least_one_sticking;
    }

    /**
     * @brief Return true if the given interface is a Signorini 
     * contact interface.
     * 
     * Argument:
     * <ul>
     *  <li> i: the interface ID.
     * </ul>
     */
    inline bool isSignorini(local_ordinal_type i)
    {
        return signorini(i) == 1;
    }

    /**
     * @brief Return true if the given interface is not a Signorini 
     * contact interface.
     * 
     * Argument:
     * <ul>
     *  <li> i: the interface ID.
     * </ul>
     */
    inline bool isNotSignorini(local_ordinal_type i)
    {
        return signorini(i) == 0;
    }

    /**
     * @brief Return true if at least one contact interface is a
     * Signorini contact interface.
     */
    inline bool isSignorini()
    {
        for (local_ordinal_type i = 0; i < contacts_number; ++i)
            if (isSignorini(i))
                return true;
        return false;
    }

    /**
     * @brief Return true if all the contact interfaces are not
     * Signorini contact interfaces.
     */
    inline bool isNotSignorini()
    {
        for (local_ordinal_type i = 0; i < contacts_number; ++i)
            if (isSignorini(i))
                return false;
        return true;
    }

    /**
     * @brief Return the number of Dirichlet nodes on the slave 
     * side for a given interface.
     * 
     * Argument:
     * <ul>
     *  <li> i: the interface ID.
     * </ul>
     */
    inline local_ordinal_type slaveDirichletSize(local_ordinal_type i)
    {
        return contacts_dirichlet(i, 0);
    }

    /**
     * @brief Return the multiplicative scale used to multiply
     * the normals of a given interface.
     * 
     * Argument:
     * <ul>
     *  <li> i: the interface ID.
     * </ul>
     */
    inline double getMortarScale(local_ordinal_type i)
    {
        return contacts_scale(i);
    }

    /**
     * @brief Return true if the node of global ID j is 
     * a slave node of the interface i.
     * 
     * Arguments:
     * <ul>
     *  <li> i: the interface ID,
     *  <li> j: the global ID of the tested node.
     * </ul>
     */
    bool isASlaveNode(local_ordinal_type i, global_ordinal_type j)
    {
        local_ordinal_type size = getSlaveNodesSize(i);
        if (size == 0)
            return false;
        for (auto j2 = 0; j2 < size; ++j2)
            if (j == getSlaveNode(i, j2))
                return true;
        return false;
    }

    /**
     * @brief Return the interface ID associated to a slave node
     * of global ID j.
     * 
     * Argument:
     * <ul>
     *  <li> j: the global ID of the node.
     * </ul>
     */
    local_ordinal_type getInterfaceOfSlaveNode(local_ordinal_type j)
    {
        for (local_ordinal_type i = 0; i < contacts_number; ++i)
            if (isASlaveNode(i, j))
                return i;
        return 0;
    }

    /**
     * @brief Return true if the slave node of global ID j 
     * is a Dirichlet node on the interface i.
     * 
     * Arguments:
     * <ul>
     *  <li> i: the interface ID,
     *  <li> j: the global ID of the node.
     * </ul>
     */
    bool isASlaveDirichletNode(local_ordinal_type i, local_ordinal_type j)
    {
        local_ordinal_type size = slaveDirichletSize(i);
        if (size == 0)
            return false;
        for (auto j2 = 0; j2 < size; ++j2)
            if (j == contacts_dirichlet(i, j2 + 1))
                return true;
        return false;
    }

    /**
     * @brief Return true if the Lagrange multiplier associated to the
     * slave node of global ID j on the interface i is initially active (closed contact).
     * 
     * Arguments:
     * <ul>
     *  <li> i: the interface ID,
     *  <li> j: the global ID of the node.
     * </ul>
     */
    bool isNodeInitiallyClosed(local_ordinal_type i, local_ordinal_type j)
    {
        local_ordinal_type size = initially_closed_nodes(i, 0);
        if (size == 0)
            return false;
        for (auto j2 = 0; j2 < size; ++j2)
            if (j == initially_closed_nodes(i, j2 + 1))
                return true;
        return false;
    }
};

}; // namespace katoptron

#endif //KATOPTRON_CONTACTSLIST_H
