#!/usr/bin/env python3
# -*- coding: utf-8; -*-

import katoptron as m
import tbox
import tbox.gmsh as gmsh
import katoptron.utilities as u
from fwk.testing import *


def main():
    comm, rank, size = u.import_MPI()

    from PyTrilinos import Teuchos

    import shlex
    import subprocess
    import os
    import numpy as np

    geo_name = '1_cube.geo'
    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    msh, mesh_name, partitioned_mesh_name = u.read_mesh_MPI(
        geo_name, file_dir, work_dir, comm, rank, size)

    ensemble_size = 1

    pbl = m.Problem(msh, comm)

    E = 1. * np.ones(ensemble_size)
    nu = 0.2 * np.ones(ensemble_size)
    k = 0.114
    d = 0.

    hf = 1 * np.ones(ensemble_size)
    zero = np.zeros(ensemble_size)
    m.Medium(pbl, "Body 1", "test", E, nu, k, d)

    num_random_variables = 10

    rand = np.random.normal(0, 1, 10*ensemble_size)

    for i in range(0, num_random_variables):
        pbl.add(m.RandomVariable(rand[i*ensemble_size:(i+1)*ensemble_size]))

    randomParams = Teuchos.ParameterList()

    randomParams['Is random'] = True
    randomParams['Begin X'] = 0.
    randomParams['Begin Y'] = 0.
    randomParams['Begin Z'] = 0.
    randomParams['End X'] = 10.
    randomParams['End Y'] = 10.
    randomParams['End Z'] = 10.
    randomParams['Number random variables'] = num_random_variables
    randomParams['Correlation length X'] = 1.
    randomParams['Correlation length Y'] = 2.
    randomParams['Correlation length Z'] = 1.

    randomParams['Mean'] = (1. / (2.*(1.+0.2)))
    randomParams['Dispersion level'] = 0.1

    m.Dirichlet(pbl, "Surf 1 1", "Clamped", 1, 0.,
                1, 0., 1, 0., 1, 10., ensemble_size)
    m.Neumann(pbl, "Surf 1 2", "heat fluxes", 0, zero,
              0, zero, 0, zero, 1, hf, ensemble_size)

    dx = np.zeros(ensemble_size)
    dy = 2. * np.ones(ensemble_size)
    dz = np.zeros(ensemble_size)
    dT = 10. * np.ones(ensemble_size)

    solverList = Teuchos.ParameterList()
    solverList['Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 1000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    solverList['randomParams'] = randomParams

    solverList['Use preconditioner'] = True

    solverList['type'] = "BlockGmres"
    solverList['Maximum active set iteration'] = 1

    solverList['Print Teuchos timers'] = True

    mueluParams = Teuchos.ParameterList()
    mueluParams['number of equations'] = 4
    mueluParams['max levels'] = 1
    mueluParams['smoother: type'] = "RELAXATION"

    mueluParamsSub2 = Teuchos.ParameterList()
    mueluParamsSub2['relaxation: type'] = "Symmetric Gauss-Seidel"
    mueluParamsSub2['relaxation: sweeps'] = 1
    mueluParamsSub2['relaxation: damping factor'] = 0.9

    mueluParams['verbosity'] = "low"

    mueluParams['smoother: params'] = mueluParamsSub2

    solverList['mueluParams'] = mueluParams
    solverList['Write txt files'] = True

    nThreads = u.Initialize_Kokkos()

    slv = m.IterativeSolver(pbl, solverList, 4, ensemble_size)

    slv.start()

    u.Finalize_Kokkos()

    tag_name = "Measure point 1"
    which_pos = 0
    which_dof = 3
    DOFperNode = 4
    name_sol = "x_mm.txt"

    pos_1, sol_1 = u.extract(msh, tag_name, DOFperNode,
                             which_pos, which_dof, ensemble_size, name_sol)

    tag_name = "Measure point 2"

    pos_2, sol_2 = u.extract(msh, tag_name, DOFperNode,
                             which_pos, which_dof, ensemble_size, name_sol)

    tests = CTests()
    measured_T_1 = sol_1[0]
    measured_T_2 = sol_2[0]
    tests.add(CTest('Temperature 1', measured_T_1, 97.7193, 5e-5))
    tests.add(CTest('Temperature 2', measured_T_2, 10., 5e-5))
    tests.run()


if __name__ == "__main__":
    main()
