/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WELEMENT_H
#define WELEMENT_H

#include "tbox.h"
#include "wObject.h"
#include <vector>
#include <Eigen/Dense>

namespace tbox
{

/**
 * @brief element types (from gmsh)
 * @authors Romain Boman
 */
enum class ELTYPE
{
    UNKNOWN = 0,
    LINE2 = 1,
    TRI3 = 2,
    QUAD4 = 3,
    TETRA4 = 4,
    HEX8 = 5,
    POINT1 = 15
};
#ifndef SWIG
TBOX_API std::ostream &operator<<(std::ostream &out, ELTYPE const &obj);
#endif

enum class MATTYPE
{
    MAT_K,
    MAT_M,
    MAT_S
};
#ifndef SWIG
TBOX_API std::ostream &operator<<(std::ostream &out, MATTYPE const &obj);
#endif

/**
 * @brief Base finite element class
 * @authors Romain Boman, Kim Liegeois, Adrien Crovato
 */
class TBOX_API Element : public fwk::wObject
{
protected:
    int order; ///< order of integration (Gauss quadrature)
public:
    int no;                    ///< label
    Tag *ptag;                 ///< physical tag   (group)
    Tag *etag;                 ///< elementary tag (geometry)
    std::vector<Tag *> parts;  ///< partitions
    std::vector<Node *> nodes; ///< nodes

#ifndef SWIG
    Element(int n, Tag *_ptag, Tag *_etag, std::vector<Tag *> &_parts, std::vector<Node *> &nods);
    virtual ~Element() {}
#endif
    virtual ELTYPE type() const
    {
        return ELTYPE::UNKNOWN;
    }

    // test functions
    void buildK_test();
    void buildM_test();
    void buildS_test();
    void builds_test();

#ifndef SWIG
    virtual Eigen::MatrixXd gp2node(Eigen::MatrixXd &values_at_gp);
    virtual double buildJ(size_t k, Eigen::MatrixXd &J, Eigen::MatrixXd &iJ) const;
    virtual double buildJ(size_t k) const;
    virtual void buildGradientJ(size_t k, std::vector<double> &dDetJ, std::vector<Eigen::MatrixXd> &dJ) const;
    virtual void buildGradientJ(size_t k, std::vector<double> &dDetJ) const;
    virtual std::vector<double> buildGradientV(int dim) const;
    //--------------------
    virtual void write(std::ostream &out) const override;
    virtual Eigen::MatrixXd buildK(std::vector<double> const &u, Fct0 const &fct);
    virtual Eigen::MatrixXd buildK(std::vector<double> const &u, Fct2 const &fct, bool fake);
    //---------------------
    virtual Eigen::MatrixXd buildK_mechanical(Eigen::MatrixXd const &H);
    virtual Eigen::MatrixXd buildS_thermomechanical(Eigen::MatrixXd const &D);
    virtual Eigen::MatrixXd buildL_thermal(Eigen::MatrixXd const &C);
    virtual void strain_stress(Eigen::MatrixXd &Strain, Eigen::MatrixXd &Stress, Eigen::MatrixXd const &H, std::vector<double> const &u);
    virtual Eigen::Vector3d normal() const;
    virtual std::vector<Eigen::Vector3d> gradientNormal() const;
    virtual std::vector<Eigen::Vector3d> computeTangents() const;
    virtual std::vector<Eigen::Vector3d> buildTangents(size_t k) const;
    //---------------------
    virtual Eigen::MatrixXd buildM();
    virtual Eigen::MatrixXd buildS();
    virtual Eigen::VectorXd builds(std::vector<double> const &u, Fct0 const &f);
    virtual Eigen::VectorXd builds(std::vector<double> const &u, Fct1 const &f);
    virtual Eigen::MatrixXd build(MATTYPE type);
    virtual Eigen::VectorXd computeGrad(std::vector<double> const &u, size_t k) const;
    virtual double computeGrad2(std::vector<double> const &u, size_t k) const;
    virtual Eigen::VectorXd computeqV(std::vector<double> const &u, Fct2 const &fct);
    virtual Eigen::VectorXd computeqint(std::vector<double> const &u, Fct2 const &fct);
    virtual double computeV(std::vector<double> const &u, Fct0 const &fct);
    virtual Eigen::MatrixXd computeV(std::vector<double> const &u, Fct2 const &fct);
    virtual double computeV();

    virtual Cache &getVCache() const = 0;
    virtual Mem &getVMem() const;
#endif
};

} // namespace tbox

#endif //WELEMENT_H
