/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wLine2.h"
#include "wSfLine2.h"
#include "wGaussLine2.h"
#include "wCacheLine2.h"
#include "wMemLine2.h"
#include "wNode.h"
#include "wFct0.h"
#include "wFct1.h"

using namespace tbox;

Line2::Line2(int n, Tag *_ptag, Tag *_etag,
             std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemLine2(*this, static_cast<int>(getCache(order).gauss.getN()));
}
Line2::~Line2()
{
    if (pmem)
        delete pmem;
}

/**
 * @brief Return private memory handling Jacobian for this Line2 element
 */
Mem &Line2::getVMem() const
{
    return getMem();
}

/**
 * @brief Return cache holding shape functions and Gauss points for all Line2 elements
 */
Cache &Line2::getVCache() const
{
    return getCache(order);
}

/**
 * @brief Compute element tangent vectors
 */
std::vector<Eigen::Vector3d> Line2::computeTangents() const
{
    return std::vector<Eigen::Vector3d>{nodes[1]->pos - nodes[0]->pos};
}

/**
 * @brief Build element tangent vectors using shape functions at Gauss point k
 */
std::vector<Eigen::Vector3d> Line2::buildTangents(size_t k) const
{
    Eigen::MatrixXd const &dff = getCache(order).getDsf(k);

    std::vector<Eigen::Vector3d> t(1, Eigen::Vector3d::Zero());
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        t[0] += (*it)->pos * dff(0, i);
    }
    return t;
}

/**
 * @brief Compute element unit normal vector
 */
Eigen::Vector3d Line2::normal() const
{
    Eigen::Vector3d t = this->computeTangents()[0].normalized();
    return Eigen::Vector3d(-t(1), t(0), 0.);
}

/**
 * @brief Compute gradient of element unit normal vector with respect to each node coordinate
 */
std::vector<Eigen::Vector3d> Line2::gradientNormal() const
{
    // Compute normal and norm
    Eigen::Vector3d t = this->computeTangents()[0];
    Eigen::Vector3d n = Eigen::Vector3d(-t(1), t(0), 0.);
    double nn = n.norm();

    // Compute gradients
    Eigen::Matrix3d inn = (Eigen::Matrix3d::Identity() - n * n.transpose() / (nn * nn)) / nn; // gradient of n/|n|
    std::vector<Eigen::Vector3d> dn(2 * 2);                                                   // 2 nodes and 2 dimensions
    dn[0] = inn * Eigen::Vector3d(0, -1, 0);
    dn[1] = inn * Eigen::Vector3d(1, 0, 0);
    dn[2] = inn * Eigen::Vector3d(0, 1, 0);
    dn[3] = inn * Eigen::Vector3d(-1, 0, 0); // outer gradient times gradient of t (pos) wtr x
    return dn;
}

/**
 * @brief Compute Jacobian determinant at Gauss point k
 *
 * detJ = norm(x_i * dN_i)
 */
double Line2::buildJ(size_t k) const
{
    return this->buildTangents(k)[0].norm();
}

/**
 * @brief Compute Jacobian determinant gradient with respect to each node coordinate, at Gauss point k
 *
 * dDetJ[l = i * 2 + j] = detJ_l * dt0_l;
 */
void Line2::buildGradientJ(size_t k, std::vector<double> &dDetJ) const
{
    Eigen::MatrixXd const &dsf = getCache(order).getDsf(k);

    // Tangent vectors and determinant
    Eigen::Vector3d detJ = this->buildTangents(k)[0].normalized();

    // Jacobian gradient
    dDetJ.resize(2 * 2); // 2 nodes and 2 dimensions (x, y)
    for (size_t i = 0; i < 2; ++i)
    {
        for (size_t j = 0; j < 2; ++j)
        {
            Eigen::Vector3d dt0 = Eigen::Vector3d::Zero();
            dt0(j) = dsf(0, i); // retain sf gradients associated to ith node for dimension j
            dDetJ[i * 2 + j] = detJ.dot(dt0);
        }
    }
}

/**
 * @brief Compute volume gradient with respect to each node coordinate
 *
 * dV[i] = w_k * dDetJ[i]_k
 */
std::vector<double> Line2::buildGradientV(int dim) const
{
    GaussLine2 &gauss = getCache(order).gauss;
    MemLine2 &mem = getMem();

    // Gradient of volume
    std::vector<double> dVol(2 * dim, 0.); // 2 nodes
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        double w = gauss.getW(k);                       // GP weight
        std::vector<double> dDetJ = mem.getGradDetJ(k); // gradient of Jacobian determinant
        for (size_t i = 0; i < dDetJ.size(); ++i)
            dVol[i] += w * dDetJ[i];
    }
    return dVol;
}

/**
 * @brief Compute flux vector
 *
 * s_i(2) = int f * N_i dV
 */
Eigen::VectorXd Line2::builds(std::vector<double> const &u, Fct0 const &f)
{
    CacheLine2 &cache = getCache(order);
    GaussLine2 &gauss = cache.gauss;
    MemLine2 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd s = Eigen::Vector2d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
        s += cache.getSf(k) * f.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    return s;
}

/**
 * @brief Compute flux vector
 *
 * s_i(2) = int f_j*n_j * N_i dV
 */
Eigen::VectorXd Line2::builds(std::vector<double> const &u, Fct1 const &f)
{
    CacheLine2 &cache = getCache(order);
    GaussLine2 &gauss = cache.gauss;
    MemLine2 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd s = Eigen::Vector2d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
        s += cache.getSf(k) * (f.eval(*this, u, k).dot(normal())) * gauss.getW(k) * mem.getDetJ(k);
    return s;
}

/**
 * @brief Compute square of gradient along element
 *
 * grad = (Delta_U / Delta_L)^2
 */
double Line2::computeGrad2(std::vector<double> const &u, size_t k) const
{
    double dUdL = (u[nodes[1]->row] - u[nodes[0]->row]) / getMem().getVol(); // V = length of the element
    return dUdL * dUdL;
}

/**
 * @brief Compute intergal of scalar function
 *
 * I = int f dV
 */
double Line2::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheLine2 &cache = getCache(order);
    GaussLine2 &gauss = cache.gauss;
    MemLine2 &mem = getMem();

    // Gauss integration
    double V = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        V += fct.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    return V;
}

/**
 * @brief Return the element length
 */
double Line2::computeV()
{
    return getMem().getVol();
}
