/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WMEM_H
#define WMEM_H

#include "tbox.h"
#include <vector>
#include <Eigen/Dense>

#ifndef SWIG

namespace tbox
{

/**
 * @brief Base class for element private memory
 * @authors Romain Boman, Adrien Crovato
 * @todo Remove capability to getJ and getJinv when not volume element
 * @todo Split values and gradients in 2 different Mem?
 */
class TBOX_API Mem
{
protected:
    // Flag
    bool valid;  ///< validity flag (true if memory has been updated once)
    bool validG; ///< validity flag for gradients (true if gradients have been initialized)

    // Element value
    size_t ngp;         ///< number of Gauss points
    double vol;         ///< element volume
    Eigen::Vector3d cg; ///< element center of gravity

    // GP values
    std::vector<double> detJs;          ///< determinant of Jacobian matrix
    std::vector<Eigen::MatrixXd> Js;    ///< Jacobian matrix
    std::vector<Eigen::MatrixXd> Jinvs; ///< inverse of Jacobian matrix

    // Gradients
    std::vector<double> gradVol;                      ///< gradient of volume wrt to nodes position
    std::vector<std::vector<double>> gradDetJs;       ///< gradient of Jacobian determinant wrt to nodes position
    std::vector<std::vector<Eigen::MatrixXd>> gradJs; ///< gradient of Jacobian matrix wrt to nodes position

    // Functions
    void checkJac(Element const &e, double detJ) const;
    virtual void updateGradients(bool sameDim);

public:
    explicit Mem(size_t _ngp);
    virtual ~Mem() {}

    inline bool isValid();
    inline bool isGradientValid();
    inline double getDetJ(size_t k);
    inline Eigen::MatrixXd const &getJ(size_t k);
    inline Eigen::MatrixXd const &getJinv(size_t k);
    inline double getVol();
    inline Eigen::Vector3d const &getCg();
    inline std::vector<double> const &getGradDetJ(size_t k);
    inline std::vector<Eigen::MatrixXd> const &getGradJ(size_t k);
    inline std::vector<double> const &getGradVol();

    void initGradients(bool sameDim);
    virtual void update(bool sameDim);
};

#include "wMem.inl"

} // namespace tbox

#endif

#endif //WMEM_H
