/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wMemHex8.h"
#include "wNode.h"

using namespace tbox;

MemHex8::MemHex8(Hex8 &_e, size_t _ngp) : Mem(_ngp), e(_e)
{
}

/**
 * @brief Update Hex8 values
 */
void MemHex8::update(bool sameDim)
{
    // Jacobian
    if (!sameDim)
        throw std::runtime_error("MemHex8::update(): the problem must always have the same dimension as Hex8!\n");
    // Hex8 is always of same dimension as Problem (volume element)
    for (int k = 0; k < ngp; ++k)
    {
        detJs[k] = e.buildJ(k, Js[k], Jinvs[k]);
        checkJac(e, detJs[k]);
    }
    // CG
    cg = Eigen::Vector3d::Zero();
    for (auto n : e.nodes)
        cg += n->pos;
    cg /= e.nodes.size();
    // Memory is now valid and the volume can be computed
    valid = true;
    vol = e.Element::computeV();

    // Update gradient if they are needed
    if (validG)
        this->updateGradients(sameDim);
}

/**
 * @brief Update Hex8 gradients
 */
void MemHex8::updateGradients(bool sameDim)
{
    // Jacobian
    if (!sameDim)
        throw std::runtime_error("MemHex8::updateGradients(): the problem must always have the same dimension as Hex8!\n");
    // Hex8 is always of same dimension as Problem (volume element)
    for (int k = 0; k < ngp; ++k)
        e.buildGradientJ(k, gradDetJs[k], gradJs[k]);
    // Volume
    validG = true;
    gradVol = e.buildGradientV(3);
}
