/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wMshCrack.h"
#include "wMshData.h"
#include "wGroup.h"
#include "wGroups.h"
#include "wTag.h"
#include "wNode.h"
#include "wElement.h"
#include "wLine2.h"
#include "wTri3.h"

using namespace tbox;

MshCrack::MshCrack(std::shared_ptr<MshData> _msh, int _nDim) : wSharedObject(), msh(_msh), nDim(_nDim)
{
    hasCrack = true; // so that we prevent automatic data structure creation if addCrack() is not called
    crckGrp = nullptr;
    exclGrp = nullptr;
}
MshCrack::~MshCrack()
{
    // Delete temporary groups
    delete crckGrp;
    delete exclGrp;
    for (size_t i = 0; i < bndGrps.size(); ++i)
        delete bndGrps[i];
    for (size_t i = 0; i < bndGrp.size(); ++i)
        delete bndGrp[i];
    std::cout << "~MshCrack()\n";
}

/**
 * @brief Create temporary group for crack
 */
void MshCrack::setCrack(std::string const &crckName)
{
    // Check if a crack needs to be made
    auto it = msh->ntags.find(crckName + '_');
    if (it != msh->ntags.end())
        std::cout << "Groups " << crckName << " and " << crckName << "_ are already existing. MshCrack will not be run!\n";
    else
    {
        hasCrack = false;
        crckGrp = new Group(msh, crckName);
    }
}

/**
 * @brief Create temporary group for excluded nodes on crack
 */
void MshCrack::setExcluded(std::string const &exclName)
{
    if (!hasCrack)
        exclGrp = new Group(msh, exclName);
}

/**
 * @brief Create temporary groups for crack boundaries
 */
void MshCrack::addBoundaries(std::vector<std::string> const &bndName)
{
    if (!hasCrack)
    {
        for (auto name : bndName)
        {
            // Check if the boundaries are explicitely separated...
            try
            {
                Groups *gs = new Groups(msh, std::vector<std::string>{name, name + '_'});
                bndGrps.push_back(gs);
            }
            // ... if not, an algorithm will later try to to separate the elements according to the side of the crack they are touching
            catch (const std::out_of_range &)
            {
                Group *g = new Group(msh, name);
                bndGrp.push_back(g);
            }
        }
    }
}

/**
 * @brief Create the crack and save the mesh
 */
void MshCrack::run()
{
    if (hasCrack)
        std::cout << "MshCrack not run!\n";
    else
    {
        // Check that crack is bounded
        if (bndGrps.empty() && bndGrp.empty())
            throw std::runtime_error("MshCrack::run crack has no boundary (use MshCrack::addBoundary()!\n");
        // Open the crack and modify data structure
        openCrack();
        modifyBoundaries_();
        modifyBoundaries();
        std::cout << "Crack added on " << *(crckGrp->tag) << "(" << crckGrp->tag->name << "_ created)" << std::endl;
    }
}

/**
 * @brief Duplicate the nodes and elements on the crack and modify the mesh accordingly
 */
void MshCrack::openCrack()
{
    // Collect the nodes to be duplicated
    std::vector<Node *> nodes;
    for (auto e : crckGrp->tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);
    std::sort(nodes.begin(), nodes.end());
    auto it = std::unique(nodes.begin(), nodes.end());
    nodes.resize(std::distance(nodes.begin(), it));
    // Remove the nodes to be excluded
    if (exclGrp != nullptr)
    {
        // Get junction nodes
        std::vector<Node *> excl;
        for (auto e : exclGrp->tag->elems)
            for (auto n : e->nodes)
                excl.push_back(n);
        std::sort(excl.begin(), excl.end());
        it = std::unique(excl.begin(), excl.end());
        excl.resize(std::distance(excl.begin(), it));
        // Find the intersection and remove
        std::vector<Node *>::iterator endRange;
        endRange = std::set_difference(nodes.begin(), nodes.end(),
                                       excl.begin(), excl.end(),
                                       nodes.begin());
        nodes.erase(endRange, nodes.end());
    }
    // Duplicate the nodes, modify the mesh and make a map
    size_t i = msh->nodes.size();
    for (auto n : nodes)
    {
        Node *newNodep = new Node(static_cast<int>(i + 1), n->pos, static_cast<int>(i));
        msh->nodes.push_back(newNodep);
        nodMap[n] = newNodep;
        i++;
    }

    // Create new tag and modify the mesh
    Tag *tagp = new Tag(static_cast<int>(msh->ptags.rbegin()->first + 1), crckGrp->tag->name + "_", nDim - 1);
    Tag *tage = new Tag(static_cast<int>(msh->etags.rbegin()->first + 1), crckGrp->tag->name + "_", nDim - 1);
    msh->ntags[crckGrp->tag->name + "_"] = tagp;
    msh->ptags[static_cast<int>(msh->ptags.rbegin()->first + 1)] = tagp;
    msh->etags[static_cast<int>(msh->etags.rbegin()->first + 1)] = tage;
    // Duplicate elements and modify the mesh/tag
    i = msh->elems.size() + 1;
    for (auto e : crckGrp->tag->elems)
    {
        std::vector<Node *> nods(e->nodes.size());
        for (size_t j = 0; j < nods.size(); ++j)
        {
            try
            {
                nods[j] = nodMap.at(e->nodes[e->nodes.size() - j - 1]);
            }
            catch (const std::out_of_range &)
            {
                nods[j] = e->nodes[e->nodes.size() - j - 1];
            }
        }
        if (e->type() == ELTYPE::LINE2 && nDim == 2)
            msh->elems.push_back(new Line2(static_cast<int>(i), tagp, tage, e->parts, nods));
        else if (e->type() == ELTYPE::TRI3 && nDim == 3)
            msh->elems.push_back(new Tri3(static_cast<int>(i), tagp, tage, e->parts, nods));
        else
        {
            std::stringstream err;
            err << "Msh::Crack: could not create element of type " << e->type() << " for dimension " << nDim << "!\n";
            throw std::runtime_error(err.str());
        }
        i++;
    }
}

/**
 * @brief Swap the elements nodes on one side of the crack and merge physical tags
 */
void MshCrack::modifyBoundaries_()
{
    for (auto gs : bndGrps)
    {
        // Modify element nodes touching the crack
        for (auto e : gs->groups[1]->tag->elems)
            swapNodes(e);
        // Merge tags
        Tag *refTag = gs->groups[0]->tag;
        Tag *oldTag = gs->groups[1]->tag;
        // Modify element tags
        for (auto e : gs->groups[1]->tag->elems)
            e->ptag = refTag;
        // Append elements to new tag
        refTag->elems.insert(refTag->elems.end(), oldTag->elems.begin(), oldTag->elems.end());
        // Delete old tag and remove from mesh maps
        msh->ptags.erase(oldTag->no);
        msh->ntags.erase(oldTag->name);
        delete oldTag;
    }
}

/**
 * @brief Detect the elements on one side of the crack and swap their nodes
 */
void MshCrack::modifyBoundaries()
{
    // compute crack CGs and normals
    std::vector<Eigen::Vector3d> crkCG(crckGrp->tag->elems.size()), crkN(crckGrp->tag->elems.size());
    for (size_t i = 0; i < crckGrp->tag->elems.size(); ++i)
    {
        crkCG[i] = Eigen::Vector3d::Zero();
        for (size_t j = 0; j < crckGrp->tag->elems[i]->nodes.size(); ++j)
            crkCG[i] += crckGrp->tag->elems[i]->nodes[j]->pos;
        crkCG[i] /= crckGrp->tag->elems[i]->nodes.size();
        crkN[i] = crckGrp->tag->elems[i]->normal();
    }
    // find elements on the opposite side of the crack and swap their nodes
    for (auto g : bndGrp)
    {
        // find elements that have at least one node on the crack and compute their CG
        std::map<Element *, Eigen::Vector3d> onCrk;
        for (auto e : g->tag->elems)
        {
            for (auto n : e->nodes)
            {
                if (nodMap.find(n) != nodMap.end())
                {
                    Eigen::Vector3d cg(0, 0, 0);
                    for (auto n_ : e->nodes)
                        cg += n_->pos;
                    cg /= e->nodes.size();
                    onCrk[e] = cg;
                    break;
                }
            }
        }
        // find closest element, check side and swap nodes
        for (auto p : onCrk)
        {
            // find mininal distance bewteen CGs
            size_t idx = 0;
            std::vector<Eigen::Vector3d> deltaCG(crkCG.size());
            for (size_t i = 0; i < crkCG.size(); ++i)
            {
                deltaCG[i] = p.second - crkCG[i];
                if (deltaCG[i].norm() < deltaCG[idx].norm())
                    idx = i;
            }
            // check sign of dot product
            if (deltaCG[idx].dot(crkN[idx]) < 0)
                swapNodes(p.first);
        }
    }
}

/**
 * @brief Replace nodes of elements on opposite side of crack by opposite nodes
 */
void MshCrack::swapNodes(Element *e)
{
    for (size_t i = 0; i < e->nodes.size(); ++i)
    {
        try
        {
            e->nodes[i] = nodMap.at(e->nodes[i]);
        }
        catch (const std::out_of_range &)
        {
            //std::cout << e->nodes[i]->no << "not found in map!\n";
        }
    }
}

void MshCrack::write(std::ostream &out) const
{
    out << "MshCrack on " << msh << std::endl;
}