/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wMshDeform.h"
#include "wMshData.h"
#include "wGroup.h"
#include "wGroups.h"
#include "wTag.h"
#include "wNode.h"
#include "wElement.h"
#include "wMem.h"
#include "wGmres.h"

#include <deque>

//#include <tbb/task_scheduler_init.h>
//#include <tbb/tbb.h>
//#include <tbb/parallel_do.h>
//#include <tbb/spin_mutex.h>

using namespace tbox;

MshDeform::MshDeform(std::shared_ptr<MshData> _msh, size_t _nDim) : wSharedObject(), nDim(_nDim), field(false), fixed(false), moving(false), msh(_msh)
{
    // Check problem dimension
    if (nDim != 2 && nDim != 3)
        throw std::runtime_error("tbox::MshDeform: nDim should be 2 or 3!\n");
    // Default
    fldGrp = nullptr;
    symBndGrp = nullptr;

    // Initialize multithreading
    nthreads = 1;
    // Get mesh size
    mshSize = msh->nodes.size();
}

/**
 * @brief Set internal field
 */
void MshDeform::setField(std::string const &fldName)
{
    // Create temporary field group
    fldGrp = new Group(msh, fldName);
    field = true;
}

/**
 * @brief Add fixed boundaries
 */
void MshDeform::addFixed(std::vector<std::string> const &fxdBndName)
{
    // Create temporary boundary groups
    for (size_t i = 0; i < fxdBndName.size(); ++i)
    {
        Group *g = new Group(msh, fxdBndName[i]);
        fxdBndGrp.push_back(g);
    }
    fixed = true;
}

/**
 * @brief Add moving boundaries
 */
void MshDeform::addMoving(std::vector<std::string> const &movBndName)
{
    // Create temporary boundary groups
    for (size_t i = 0; i < movBndName.size(); ++i)
    {
        Group *g = new Group(msh, movBndName[i]);
        movBndGrp.push_back(g);
    }
    moving = true;
}

/**
 * @brief Set symmetry boundary
 */
void MshDeform::setSymmetry(std::string const &symBndName, size_t _blck)
{
    // Get blocked direction
    blck = _blck;
    if (blck != 0 && blck != 1 && blck != 2)
        throw std::runtime_error("tbox::MshDeform::setSymmetry: blck should be 0, 1 or 2 (for x, y and z direction resp.)!\n");
    // Create temporary boundary groups
    symBndGrp = new Group(msh, symBndName);
}

/**
 * @brief Add internal boundaries
 */
void MshDeform::addInternal(std::vector<std::string> const &intBndName)
{
    // Create temporary internal link
    if (intBndName.size() != 2)
        throw std::runtime_error("MshDeform: input intBndName must be a pair of internal boundaries!\n");
    Groups *gs = new Groups(msh, intBndName);
    intBndGrps.push_back(gs);
}

/**
 * @brief Initialize data structure
 */
void MshDeform::initialize()
{
    // Sanity checks
    if (!field)
        throw std::runtime_error("MshDeform: field has not been found! Use setField() to add a field.\n");
    else if (!fixed)
        throw std::runtime_error("MshDeform: fixed boundaries have not been found! Use addFixed() to add at least one boundary.\n");
    else if (!moving)
        throw std::runtime_error("MshDeform: moving boundaries have not been found! Use addMoving() to add at least one boundary.\n");

    // Store the field elements
    for (auto e : fldGrp->tag->elems)
        fldElems.push_back(e);

    // Create vector of fixed boundaries nodes
    for (auto bnd : fxdBndGrp)
        for (auto e : bnd->tag->elems)
            for (auto n : e->nodes)
                fxdBndNodes.push_back(n);
    uniqueSort(fxdBndNodes);

    // Create vector of moving boundaries nodes
    for (auto bnd : movBndGrp)
    {
        for (auto e : bnd->tag->elems)
        {
            movBndElems.push_back(e);
            for (auto n : e->nodes)
                movBndNodes.push_back(n);
        }
    }
    uniqueSort(movBndNodes);

    // Create vector of symmetry boundary nodes
    if (symBndGrp)
    {
        for (auto e : symBndGrp->tag->elems)
        {
            symBndElems.push_back(e);
            for (auto n : e->nodes)
                symBndNodes.push_back(n);
        }
        uniqueSort(symBndNodes);
    }

    // Create the nodes pair for internal boundaries
    for (auto bnds : intBndGrps)
    {
        // Temporary container for internal nodes
        std::vector<Node *> nodes0, nodes1;
        for (auto e : bnds->groups[0]->tag->elems)
        {
            intBndElems.push_back(e);
            for (auto n : e->nodes)
                nodes0.push_back(n);
        }
        for (auto e : bnds->groups[1]->tag->elems)
        {
            intBndElems.push_back(e);
            for (auto n : e->nodes)
                nodes1.push_back(n);
        }
        uniqueSort(nodes0);
        uniqueSort(nodes1);
        // Create the nodes pair
        if (nodes0.size() != nodes1.size())
            throw std::runtime_error("MshDeform: internal boudnaries must have the same number of nodes!");
        for (auto n0 : nodes0)
        {
            for (auto n1 : nodes1)
            {
                if ((n0->pos - n1->pos).norm() <= 1e-14 && n0 != n1)
                {
                    intBndPair.push_back(std::pair<Node *, Node *>(n0, n1));
                    break;
                }
            }
        }
    }

    // List of rows to build stiffness matrix
    rows.resize(nDim * mshSize);
    for (size_t i = 0; i < mshSize; ++i)
        for (size_t m = 0; m < nDim; ++m)
            rows[nDim * i + m] = nDim * msh->nodes[i]->row + m;
    // Periodic boundary condition step 1
    // -> for each node pair(a, b): node a contributions will be added to node b rows
    for (auto p : intBndPair)
        for (size_t m = 0; m < nDim; ++m)
            rows[nDim * (p.first->row) + m] = nDim * (p.second->row) + m;

    // Delete temporary
    delete fldGrp;
    for (size_t i = 0; i < fxdBndGrp.size(); ++i)
        delete fxdBndGrp[i];
    for (size_t i = 0; i < movBndGrp.size(); ++i)
        delete movBndGrp[i];
    delete symBndGrp;
    for (size_t i = 0; i < intBndGrps.size(); ++i)
        delete intBndGrps[i];
}

/**
 * @brief Store moving nodes position before deformation
 */
void MshDeform::savePos()
{
    initPos.resize(nDim * movBndNodes.size());
    for (size_t i = 0; i < movBndNodes.size(); ++i)
        for (int m = 0; m < nDim; ++m)
            initPos[nDim * i + m] = movBndNodes[i]->pos(m);
}

/**
 * @brief Build the mesh deformation matrix using linear elasticity law
 *
 * Dwight RP, Robust Mesh Deformation using the Linear Elasticity Equations.
 * Journal of Computational Fluid Dynamics 12: 401-406, 2009
 */
void MshDeform::build(Eigen::SparseMatrix<double, Eigen::RowMajor> &K)
{
    //    // Multithread
    //    tbb::spin_mutex mutex;

    // List of triplets to build stifness matrix
    std::deque<Eigen::Triplet<double>> T;

    // Build stiffness matrix
    //    tbb::parallel_do(fldElems.begin(), fldElems.end(), [&](Element* e) {
    for (auto e : fldElems)
    {
        // Hooke tensor
        Eigen::MatrixXd H;
        if (nDim == 2)
        {
            H = Eigen::Matrix3d::Zero();
            double nu = 0.;               // Poisson ratio [-1, 0.5]
            double E = 1 / e->computeV(); // Young modulus
            H(0, 0) = 1 - nu;
            H(1, 1) = 1 - nu;
            H(2, 2) = (1 - 2 * nu) / 2;
            H(0, 1) = nu;
            H(1, 0) = H(0, 1);
            H *= E / ((1 + nu) * (1 - 2 * nu));
        }
        else
        {
            H = Eigen::MatrixXd::Zero(6, 6);
            double nu = 0.;               // Poisson ratio [-1, 0.5]
            double E = 1 / e->computeV(); // Young modulus
            H(0, 0) = 1 - nu;
            H(1, 1) = 1 - nu;
            H(2, 2) = 1 - nu;
            H(3, 3) = (1 - 2 * nu) / 2;
            H(4, 4) = (1 - 2 * nu) / 2;
            H(5, 5) = (1 - 2 * nu) / 2;
            H(0, 1) = nu;
            H(1, 0) = H(0, 1);
            H(0, 2) = nu;
            H(2, 0) = H(0, 2);
            H(1, 2) = nu;
            H(2, 1) = H(1, 2);
            H *= E / ((1 + nu) * (1 - 2 * nu));
        }

        // Elementary stiffness matrix
        Eigen::MatrixXd Ke = e->buildK_mechanical(H);

        // Assembly
        //        tbb::spin_mutex::scoped_lock lock(mutex);
        for (size_t i = 0; i < e->nodes.size(); ++i)
        {
            for (int m = 0; m < nDim; ++m)
            {
                size_t rowi = nDim * (e->nodes[i]->row) + m;
                for (size_t j = 0; j < e->nodes.size(); ++j)
                {
                    for (int n = 0; n < nDim; ++n)
                    {
                        size_t rowj = nDim * (e->nodes[j]->row) + n;
                        T.push_back(Eigen::Triplet<double>(rows[rowi], rowj, Ke(nDim * i + m, nDim * j + n)));
                    }
                }
            }
        }
    }
    //    });
    // Periodic boundary condition step 2
    // -> for each node pair(a, b): write "pos_a - pos_b" = 0 on row a
    for (auto p : intBndPair)
        for (int m = 0; m < nDim; ++m)
        {
            T.push_back(Eigen::Triplet<double>(nDim * (p.first->row) + m, nDim * (p.first->row) + m, 1.));
            T.push_back(Eigen::Triplet<double>(nDim * (p.first->row) + m, nDim * (p.second->row) + m, -1.));
        }
    // Build matrix without BCs
    K.setFromTriplets(T.begin(), T.end());
    // Apply Dirichlet condition on fixed boundaries
    for (size_t i = 0; i < fxdBndNodes.size(); i++)
    {
        for (int m = 0; m < nDim; ++m)
        {
            size_t rowi = nDim * (fxdBndNodes[i]->row) + m;
            for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(K, rowi); it; ++it)
            {
                if (it.row() == it.col())
                    it.valueRef() = 1.;
                else
                    it.valueRef() = 0.;
            }
        }
    }
    // Apply Dirichlet condition on symmetry boundaries
    for (size_t i = 0; i < symBndNodes.size(); i++)
    {
        size_t rowi = nDim * (symBndNodes[i]->row) + blck;
        for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(K, rowi); it; ++it)
        {
            if (it.row() == it.col())
                it.valueRef() = 1.;
            else
                it.valueRef() = 0.;
        }
    }
    // Apply Dirichlet condition on moving boundaries
    for (size_t i = 0; i < movBndNodes.size(); i++)
    {
        for (int m = 0; m < nDim; ++m)
        {
            size_t rowi = nDim * (movBndNodes[i]->row) + m;
            for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(K, rowi); it; ++it)
            {
                if (it.row() == it.col())
                    it.valueRef() = 1.;
                else
                    it.valueRef() = 0.;
            }
        }
    }
    // Clean matrix and turn to compressed row format
    K.prune(0.);
    K.makeCompressed();
}

/**
 * @brief Build the mesh deformation force vector (B.C.) using linear elasticity law
 *
 * Dwight RP, Robust Mesh Deformation using the Linear Elasticity Equations.
 * Journal of Computational Fluid Dynamics 12: 401-406, 2009
 */
void MshDeform::build(Eigen::VectorXd &f)
{
    // Apply Dirichlet condition on moving boundaries
    for (size_t i = 0; i < movBndNodes.size(); i++)
    {
        for (int m = 0; m < nDim; ++m)
        {
            size_t rowi = nDim * (movBndNodes[i]->row) + m;
            f(rowi) = movBndNodes[i]->pos(m) - initPos[nDim * i + m];
        }
    }
}

/**
 * @brief Deform the mesh following linear elasticity law
 *
 * Dwight RP, Robust Mesh Deformation using the Linear Elasticity Equations.
 * Journal of Computational Fluid Dynamics 12: 401-406, 2009
 */
void MshDeform::deform()
{
    //    // Multithread
    //    tbb::task_scheduler_init init(nthreads);

    // Stiffness matrix
    Eigen::SparseMatrix<double, Eigen::RowMajor> K(nDim * mshSize, nDim * mshSize);
    // force (RHS) and displacement (UKN) vectors
    Eigen::VectorXd f = Eigen::VectorXd::Zero(nDim * mshSize), q = Eigen::VectorXd::Zero(nDim * mshSize);

    // Build stiffness matrix and RHS
    std::cout << "--- Deforming mesh using linear elasticiy equations ---\n"
              << "Number of threads: " << nthreads << "\n"
              << "Assembling matrices... " << std::flush;
    this->build(K);
    this->build(f);
    std::cout << "done" << std::endl;
    // Solve the SoE
    std ::cout << "Solving equations... " << std::flush;
    Gmres gmres;
    // set the preconditioner
    gmres.setFillFactor(2);
    gmres.setDropTol(1e-6);
    // set the gmres
    gmres.setTolerance(1e-8);
    gmres.setRestart(50);
    gmres.setMaxIterations(500);
    // solve
    Eigen::Map<Eigen::VectorXd> f_(f.data(), f.size()), q_(q.data(), q.size());
    gmres.compute(K, f_, q_);
    std::cout << "done (#it: " << gmres.getIterations() << ", error: " << std::scientific << gmres.getError() << std::fixed << ")" << std::endl;

    // Update position
    std ::cout << "Updating mesh... " << std::flush;
    for (auto n : msh->nodes)
    {
        for (int m = 0; m < nDim; m++)
            n->pos(m) += q(nDim * (n->row) + m);
    }
    // Reset moving boundary position
    for (auto n : movBndNodes)
    {
        for (int m = 0; m < nDim; m++)
            n->pos(m) -= q(nDim * (n->row) + m);
    }
    // Update element memory
    for (auto e : fldElems)
        e->getVMem().update(true);
    for (auto e : movBndElems)
        e->getVMem().update(false);
    for (auto e : symBndElems)
        e->getVMem().update(false);
    for (auto e : intBndElems)
        e->getVMem().update(false);
    std::cout << "done" << std::endl;
}

/**
 * @brief Sort a vector and remove duplicates
 */
void MshDeform::uniqueSort(std::vector<Node *> &nodes)
{
    std::sort(nodes.begin(), nodes.end());
    auto it = std::unique(nodes.begin(), nodes.end());
    nodes.resize(std::distance(nodes.begin(), it));
}

void MshDeform::write(std::ostream &out) const
{
    out << "MshDeform on " << msh << std::endl;
}
