/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wQuad4.h"
#include "wSfQuad4.h"
#include "wGaussQuad4.h"
#include "wCacheQuad4.h"
#include "wMemQuad4.h"
#include "wNode.h"
#include "wFct0.h"

using namespace tbox;

Quad4::Quad4(int n, Tag *_ptag, Tag *_etag,
             std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemQuad4(*this, static_cast<int>(getCache(order).gauss.getN()));
}
Quad4::~Quad4()
{
    if (pmem)
        delete pmem;
}

/**
 * @brief Return private memory handling Jacobian for this Quad4 element
 */
Mem &Quad4::getVMem() const
{
    return getMem();
}

/**
 * @brief Return cache holding shape functions and Gauss points for all Quad4 elements
 */
Cache &Quad4::getVCache() const
{
    return getCache(order);
}

/**
 * @brief Compute element tangent vectors
 */
std::vector<Eigen::Vector3d> Quad4::computeTangents() const
{
    Eigen::Vector3d x1 = nodes[0]->pos;
    Eigen::Vector3d x3 = nodes[2]->pos;
    Eigen::Vector3d x4 = nodes[3]->pos;

    return std::vector<Eigen::Vector3d>{(x3 - x1), (x4 - x1)};
}

/**
 * @brief Build element tangent vectors using shape functions at Gauss point k
 */
std::vector<Eigen::Vector3d> Quad4::buildTangents(size_t k) const
{
    Eigen::MatrixXd const &dff = getCache(order).getDsf(k);

    std::vector<Eigen::Vector3d> t(2, Eigen::Vector3d::Zero());
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        t[0] += (*it)->pos * dff(0, i);
        t[1] += (*it)->pos * dff(1, i);
    }
    return t;
}

/**
 * @brief Compute element unit normal vector
 */
Eigen::Vector3d Quad4::normal() const
{
    std::vector<Eigen::Vector3d> ts = this->computeTangents();
    return ts[0].cross(ts[1]).normalized();
}

/**
 * @brief Compute Jacobian determinant at Gauss point k
 *
 * detJ = norm(x_i dN_i,xi X x_i dN_i,eta)
 */
double Quad4::buildJ(size_t k) const
{
    std::vector<Eigen::Vector3d> ts = this->buildTangents(k);
    return ts[0].cross(ts[1]).norm();
}
/**
 * @brief Compute Jacobian matrix, inverse and determinant at Gauss point k
 *
 * J_ij(2,2) = diN_n * xj_n (n is the node)
 */
double Quad4::buildJ(size_t k, Eigen::MatrixXd &J, Eigen::MatrixXd &iJ) const
{
    Eigen::MatrixXd const &dff = getCache(order).getDsf(k);

    Eigen::Matrix2d JJ = Eigen::Matrix2d::Zero(); // temporary fixed-size matrix to efficiently compute the inverse
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
        JJ += dff.col(i) * (*it)->pos.transpose().block<1, 2>(0, 0);
    J = JJ;
    iJ = JJ.inverse();
    return JJ.determinant();
}

/**
 * @brief Compute mass matrix
 *
 * S_ij(4,4) = int N_i*N_j dV
 */
Eigen::MatrixXd Quad4::buildS()
{
    CacheQuad4 &cache = getCache(order);
    GaussQuad4 &gauss = cache.gauss;
    MemQuad4 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd S = Eigen::Matrix4d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Shape functions
        Eigen::VectorXd const &ff = cache.getSf(k);

        // Mass matrix
        S += ff * ff.transpose() * gauss.getW(k) * mem.getDetJ(k);
    }
    return S;
}

/**
 * @brief API for build methods
 */
Eigen::MatrixXd Quad4::build(MATTYPE type)
{
    switch (type)
    {
    case MATTYPE::MAT_S:
        return buildS();
        break;
    default:
        std::stringstream out;
        out << "Quad4 cannot build " << type;
        throw std::runtime_error(out.str());
    }
}

/**
 * @brief Compute intergal of scalar function
 *
 * I = int f dV
 */
double Quad4::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheQuad4 &cache = getCache(order);
    GaussQuad4 &gauss = cache.gauss;
    MemQuad4 &mem = getMem();

    // Gauss integration
    double V = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        V += fct.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    return V;
}

/**
 * @brief Return the element surface
 */
double Quad4::computeV()
{
    return getMem().getVol();
}