/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wTetra4.h"
#include "wSfTetra4.h"
#include "wGaussTetra4.h"
#include "wCacheTetra4.h"
#include "wMemTetra4.h"
#include "wNode.h"
#include "wFct0.h"

using namespace tbox;

Tetra4::Tetra4(int n, Tag *_ptag, Tag *_etag,
               std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemTetra4(*this, static_cast<int>(getCache(order).gauss.getN()));
}
Tetra4::~Tetra4()
{
    if (pmem)
        delete pmem;
}

/**
 * @brief Return private memory handling Jacobian for this Tetra4 element
 */
Mem &Tetra4::getVMem() const
{
    return getMem();
}

/**
 * @brief Return cache handling shape functions and Gauss points for all Tetra4 elements
 */
Cache &Tetra4::getVCache() const
{
    return getCache(order);
}

/**
 * @brief Compute Jacobian matrix, inverse and determinant at Gauss point k
 *
 * J_ij(3,3) = diN_n * xj_n (n is the node)
 */
double Tetra4::buildJ(size_t k, Eigen::MatrixXd &J, Eigen::MatrixXd &iJ) const
{
    Eigen::MatrixXd const &dff = getCache(order).getDsf(k);

    Eigen::Matrix3d JJ = Eigen::Matrix3d::Zero(); // temporary fixed-size matrix to efficiently compute the inverse
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
        JJ += dff.col(i) * ((*it)->pos).transpose();
    J = JJ;
    iJ = JJ.inverse();
    return JJ.determinant();
}

/**
 * @brief Compute Jacobian matrix gradient with respect to each node coordinate, at Gauss point k
 *
 * dJ[l = i * 3 + j](3,3) = dN_i(3,1) * dxy_i(1,3)/dx_l (for node i and dimension j)
 */
void Tetra4::buildGradientJ(size_t k, std::vector<double> &dDetJ, std::vector<Eigen::MatrixXd> &dJ) const
{
    MemTetra4 &mem = getMem();
    Eigen::MatrixXd const &dsf = getCache(order).getDsf(k);

    // Jacobian
    dJ.resize(4 * 3); // 4 nodes and 3 dimensions (x, y, z)
    for (size_t i = 0; i < 4; ++i)
    {
        for (size_t j = 0; j < 3; ++j)
        {
            dJ[i * 3 + j] = Eigen::Matrix3d::Zero();
            dJ[i * 3 + j].col(j) = dsf.col(i); // retain sf gradients associated to ith node for dimension j
        }
    }
    // Determinant
    double detJ = mem.getDetJ(k);
    Eigen::MatrixXd const &iJ = mem.getJinv(k);
    dDetJ.resize(4 * 3);
    for (size_t i = 0; i < dDetJ.size(); ++i)
        dDetJ[i] = detJ * (iJ * dJ[i]).trace();
}
/**
 * @brief Compute volume gradient with respect to each node coordinate
 *
 * dV[i] = w_k * dDetJ[i]_k
 */
std::vector<double> Tetra4::buildGradientV(int dim) const
{
    GaussTetra4 &gauss = getCache(order).gauss;
    MemTetra4 &mem = getMem();

    // Gradient of volume
    std::vector<double> dVol(4 * dim, 0.); // 4 nodes
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        double w = gauss.getW(k);                       // GP weight
        std::vector<double> dDetJ = mem.getGradDetJ(k); // gradient of Jacobian determinant
        for (size_t i = 0; i < dDetJ.size(); ++i)
            dVol[i] += w * dDetJ[i];
    }
    return dVol;
}

/**
 * @brief Compute mass matrix
 *
 * M_ij(4,4) = int N_i*N_j dV
 */
Eigen::MatrixXd Tetra4::buildM()
{
    CacheTetra4 &cache = getCache(order);
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd M = Eigen::Matrix4d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Shape functions
        Eigen::VectorXd const &ff = cache.getSf(k);
        M += ff * ff.transpose() * gauss.getW(k) * mem.getDetJ(k);
    }
    return M;
}

/**
 * @brief Compute stiffness matrix for scalar field
 *
 * K_ij(4,4) = int f*dN_j*dN_i dV
 */
Eigen::MatrixXd Tetra4::buildK(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTetra4 &cache = getCache(order);
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd K = Eigen::Matrix4d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian inverse and shape functions
        Eigen::Matrix3d const &J = mem.getJinv(k);
        Eigen::MatrixXd const &dff = cache.getDsf(k);

        // Elementary stiffness matrix
        K += (J * dff).transpose() * J * dff * fct.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    }
    return K;
}

/**
 * @brief Compute linear elasticity stiffness matrix
 *
 * K_ij(12,12) = int eps_k H_ijkl eps_l dV = B(12,6)H(6,6)B(6,12)
 */
Eigen::MatrixXd Tetra4::buildK_mechanical(Eigen::MatrixXd const &H)
{
    CacheTetra4 &cache = getCache(order);
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    Eigen::MatrixXd K = Eigen::MatrixXd::Zero(12, 12);
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian inverse
        Eigen::Matrix3d const &invJ = mem.getJinv(k);
        // Fill B matrix (shape functions)
        Eigen::MatrixXd const &dff = cache.getDsf(k);
        Eigen::MatrixXd B = Eigen::MatrixXd::Zero(6, 3 * nodes.size());
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            // Compute [Ni,x Ni,y, Ni_z]
            Eigen::Vector3d dN = invJ * dff.col(i);
            B(0, 3 * i) = dN(0);
            B(1, 3 * i + 1) = dN(1);
            B(2, 3 * i + 2) = dN(2);
            B(3, 3 * i + 1) = dN(2);
            B(3, 3 * i + 2) = dN(1);
            B(4, 3 * i) = dN(2);
            B(4, 3 * i + 2) = dN(0);
            B(5, 3 * i) = dN(1);
            B(5, 3 * i + 1) = dN(0);
        }

        // Compute stiffness matrix
        K += B.transpose() * H * B * gauss.getW(k) * mem.getDetJ(k);
    }
    return K;
}

/**
 * @brief Compute thermomechanical mass matrix
 */
Eigen::MatrixXd Tetra4::buildS_thermomechanical(Eigen::MatrixXd const &D)
{
    CacheTetra4 &cache = getCache(order);
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd S_e = Eigen::MatrixXd::Zero(12, 4);
    for (size_t a = 0; a < gauss.getN(); ++a)
    {
        // Jacobian and shape functions
        double detJ = mem.getDetJ(a);
        Eigen::Matrix3d const &J = mem.getJinv(a);
        Eigen::MatrixXd const &dff = cache.getDsf(a);

        // S
        for (size_t i = 0; i < nodes.size(); ++i)
            S_e.block(i * 3, 0, 3, nodes.size()) += D * J * dff.col(i) * cache.getSf(a).transpose() * gauss.getW(a) * detJ;
    }
    return S_e;
}

/**
 * @brief Compute thermal matrix
 */
Eigen::MatrixXd Tetra4::buildL_thermal(Eigen::MatrixXd const &C)
{
    CacheTetra4 &cache = getCache(order);
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd L_e = Eigen::Matrix4d::Zero();
    for (size_t a = 0; a < gauss.getN(); ++a)
    {
        // Jacobian
        double detJ = mem.getDetJ(a);
        Eigen::Matrix3d const &J = mem.getJinv(a);
        Eigen::MatrixXd const &dff = cache.getDsf(a);

        // [AC] I refactored this from an entry by entry assembly to a full matrix multiplication,
        // but as nothing tests this method, I could not check the results.
        L_e += (C * J * dff).transpose() * J * dff * gauss.getW(a) * detJ;
    }
    return L_e;
}

/**
 * @brief Compute stress and strain
 */
void Tetra4::strain_stress(Eigen::MatrixXd &Strain, Eigen::MatrixXd &Stress, Eigen::MatrixXd const &H, std::vector<double> const &u)
{
    CacheTetra4 &cache = getCache(order);
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    Strain = Stress = Eigen::Matrix3d::Zero();
    Eigen::MatrixXd u_e(3, nodes.size());
    for (size_t i = 0; i < nodes.size(); ++i)
    {
        u_e(0, i) = u[3 * (nodes[i]->row)];
        u_e(1, i) = u[3 * (nodes[i]->row) + 1];
        u_e(2, i) = u[3 * (nodes[i]->row) + 2];
    }
    for (size_t a = 0; a < gauss.getN(); ++a)
    {
        // Jacobian and strain
        Eigen::Matrix3d const &J = mem.getJinv(a);
        Eigen::MatrixXd const &dff = cache.getDsf(a);
        Eigen::Matrix3d du_d_Phi = Eigen::Matrix3d::Zero();
        for (size_t i = 0; i < nodes.size(); ++i)
            du_d_Phi += dff.col(i) * u_e.col(i).transpose();
        Strain += ((J * du_d_Phi) + (J * du_d_Phi).transpose()) / nodes.size() / 2;
    }
    // stress
    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            for (size_t k = 0; k < 3; ++k)
                for (size_t l = 0; l < 3; ++l)
                    Stress(i, j) += H(i * 3 + j, k * 3 + l) * Strain(k, l);
}

/**
 * @brief Compute gradient of a scalar field at Gauss point k
 */
Eigen::VectorXd Tetra4::computeGrad(std::vector<double> const &u, size_t k) const
{
    // Solution vector
    Eigen::Vector4d ue;
    for (size_t i = 0; i < nodes.size(); ++i)
        ue(i) = u[nodes[i]->row];
    // gradient in global axis: inv(J)_ij * d_jN_i * ue_i
    return getMem().getJinv(k) * getCache(order).getDsf(k) * ue;
}

/**
 * @brief Compute square of gradient of a scalar field at Gauss point k
 */
double Tetra4::computeGrad2(std::vector<double> const &u, size_t k) const
{
    return computeGrad(u, k).squaredNorm();
}

/**
 * @brief Compute intergal of scalar function
 *
 * I = int f dV
 */
double Tetra4::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTetra4 &cache = getCache(order);
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    // Gauss integration
    double V = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        V += fct.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    return V;
}

/**
 * @brief Return the element volume
 */
double Tetra4::computeV()
{
    return getMem().getVol();
}
