/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wTri3.h"
#include "wSfTri3.h"
#include "wGaussTri3.h"
#include "wCacheTri3.h"
#include "wMemTri3.h"
#include "wNode.h"
#include "wFct0.h"
#include "wFct1.h"
#include "wFct2.h"

using namespace tbox;

Tri3::Tri3(int n, Tag *_ptag, Tag *_etag,
           std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemTri3(*this, static_cast<int>(getCache(order).gauss.getN()));
}
Tri3::~Tri3()
{
    if (pmem)
        delete pmem;
}

/**
 * @brief Return private memory handling Jacobian for this Tri3 element
 */
Mem &Tri3::getVMem() const
{
    return getMem();
}

/**
 * @brief Return cache holding shape functions and Gauss points for all Tri3 elements
 */
Cache &Tri3::getVCache() const
{
    return getCache(order);
}

/**
 * @brief Compute element tangent vectors
 */
std::vector<Eigen::Vector3d> Tri3::computeTangents() const
{
    Eigen::Vector3d x1 = nodes[0]->pos;
    Eigen::Vector3d x2 = nodes[1]->pos;
    Eigen::Vector3d x3 = nodes[2]->pos;

    return std::vector<Eigen::Vector3d>{(x2 - x1), (x3 - x1)};
}

/**
 * @brief Build element tangent vectors using shape functions at Gauss point k
 */
std::vector<Eigen::Vector3d> Tri3::buildTangents(size_t k) const
{
    Eigen::MatrixXd const &dff = getCache(order).getDsf(k);

    std::vector<Eigen::Vector3d> t(2, Eigen::Vector3d::Zero());
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        t[0] += (*it)->pos * dff(0, i);
        t[1] += (*it)->pos * dff(1, i);
    }
    return t;
}

/**
 * @brief Compute element unit normal vector
 */
Eigen::Vector3d Tri3::normal() const
{
    std::vector<Eigen::Vector3d> ts = this->computeTangents();
    return ts[0].cross(ts[1]).normalized();
}

/**
 * @brief Compute gradient of element unit normal vector with respect to each node coordinate
 */
std::vector<Eigen::Vector3d> Tri3::gradientNormal() const
{
    // Compute normal and norm
    std::vector<Eigen::Vector3d> ts = this->computeTangents();
    Eigen::Vector3d n = ts[0].cross(ts[1]);
    double nn = n.norm();

    // Compute gradients
    Eigen::Matrix3d inn = (Eigen::Matrix3d::Identity() - n * n.transpose() / (nn * nn)) / nn; // gradient of n/|n|
    std::vector<Eigen::Vector3d> dn(3 * 3);                                                   // 3 nodes and 3 dimensions
    for (size_t i = 0; i < 3; ++i)
    {
        for (size_t j = 0; j < 3; ++j)
        {
            Eigen::Matrix3d dx = Eigen::Matrix3d::Zero();
            dx(j, i) = 1;                                                                                        // dpos_i / dx_j
            dn[i * 3 + j] = inn * ((dx.col(1) - dx.col(0)).cross(ts[1]) + (ts[0]).cross(dx.col(2) - dx.col(0))); // outer gradient times gradient of cross product
        }
    }
    return dn;
}

/**
 * @brief Compute Jacobian determinant at Gauss point k
 *
 * detJ = norm(x_i dN_i,xi X x_i dN_i,eta)
 */
double Tri3::buildJ(size_t k) const
{
    std::vector<Eigen::Vector3d> ts = this->buildTangents(k);
    return ts[0].cross(ts[1]).norm();
}
/**
 * @brief Compute Jacobian matrix, inverse and determinant at Gauss point k
 *
 * J_ij(2,2) = diN_n * xj_n (n is the node)
 */
double Tri3::buildJ(size_t k, Eigen::MatrixXd &J, Eigen::MatrixXd &iJ) const
{
    Eigen::MatrixXd const &dff = getCache(order).getDsf(k);

    Eigen::Matrix2d JJ = Eigen::Matrix2d::Zero(); // temporary fixed-size matrix to efficiently compute the inverse
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
        JJ += dff.col(i) * (*it)->pos.transpose().block<1, 2>(0, 0);
    J = JJ;
    iJ = JJ.inverse();
    return JJ.determinant();
}

/**
 * @brief Compute Jacobian determinant gradient with respect to each node coordinate, at Gauss point k
 *
 * dDetJ[l = i * 3 + j] = detJ_l * (dt0 \cross t1 + t0 \cross dt1)_l;
 */
void Tri3::buildGradientJ(size_t k, std::vector<double> &dDetJ) const
{
    Eigen::MatrixXd const &dsf = getCache(order).getDsf(k);

    // Tangent vectors and determinant
    std::vector<Eigen::Vector3d> ts = this->buildTangents(k);
    Eigen::Vector3d detJ = ts[0].cross(ts[1]).normalized();

    // Jacobian gradient
    dDetJ.resize(3 * 3); // 3 nodes and 3 dimensions (x, y, z)
    for (size_t i = 0; i < 3; ++i)
    {
        for (size_t j = 0; j < 3; ++j)
        {
            Eigen::Vector3d dt0 = Eigen::Vector3d::Zero(), dt1 = Eigen::Vector3d::Zero();
            dt0(j) = dsf(0, i);
            dt1(j) = dsf(1, i); // retain sf gradients associated to ith node for dimension j
            dDetJ[i * 3 + j] = detJ.dot(dt0.cross(ts[1]) + ts[0].cross(dt1));
        }
    }
}
/**
 * @brief Compute Jacobian matrix gradient with respect to each node coordinate, at Gauss point k
 *
 * dJ[l = i * 2 + j](2,2) = dN_i(2,1) * dxy_i(1,2)/dx_l (for node i and dimension j)
 */
void Tri3::buildGradientJ(size_t k, std::vector<double> &dDetJ, std::vector<Eigen::MatrixXd> &dJ) const
{
    MemTri3 &mem = getMem();
    Eigen::MatrixXd const &dsf = getCache(order).getDsf(k);

    // Jacobian
    dJ.resize(3 * 2); // 3 nodes and 2 dimensions (x, y)
    for (size_t i = 0; i < 3; ++i)
    {
        for (size_t j = 0; j < 2; ++j)
        {
            dJ[i * 2 + j] = Eigen::Matrix2d::Zero();
            dJ[i * 2 + j].col(j) = dsf.col(i); // retain sf gradients associated to ith node for dimension j
        }
    }
    // Determinant
    double detJ = mem.getDetJ(k);
    Eigen::MatrixXd const &iJ = mem.getJinv(k);
    dDetJ.resize(3 * 2);
    for (size_t i = 0; i < dDetJ.size(); ++i)
        dDetJ[i] = detJ * (iJ * dJ[i]).trace();
}
/**
 * @brief Compute volume gradient with respect to each node coordinate
 *
 * dV[i] = w_k * dDetJ[i]_k
 */
std::vector<double> Tri3::buildGradientV(int dim) const
{
    GaussTri3 &gauss = getCache(order).gauss;
    MemTri3 &mem = getMem();

    // Gradient of volume
    std::vector<double> dVol(3 * dim, 0.); // 3 nodes
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        double w = gauss.getW(k);                       // GP weight
        std::vector<double> dDetJ = mem.getGradDetJ(k); // gradient of Jacobian determinant
        for (size_t i = 0; i < dDetJ.size(); ++i)
            dVol[i] += w * dDetJ[i];
    }
    return dVol;
}

/**
 * @brief Compute stiffness matrix for scalar field
 *
 * K_ij(3,3) = int f*dN_j*dN_i dV
 */
Eigen::MatrixXd Tri3::buildK(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTri3 &cache = getCache(order);
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd K = Eigen::Matrix3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian inverse and shape functions
        Eigen::Matrix2d const &J = mem.getJinv(k);
        Eigen::MatrixXd const &dff = cache.getDsf(k);

        // Elementary stiffness matrix
        K += (J * dff).transpose() * J * dff * fct.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    }
    return K;
}

/**
 * @brief Compute stiffness matrix for scalar field
 *
 * K_ij(3,3) = int [f]*dN_j*dN_i dV
 */
Eigen::MatrixXd Tri3::buildK(std::vector<double> const &u, Fct2 const &fct, bool fake)
{
    CacheTri3 &cache = getCache(order);
    GaussTri3 &gauss = cache.gauss;

    Eigen::MatrixXd K = Eigen::Matrix3d::Zero();
    if (fake) // fake run - fill MPI job list
    {
        for (size_t k = 0; k < gauss.getN(); ++k)
        {
            Eigen::MatrixXd fk;
            fct.eval(*this, u, k, fk, fake);
        }
    }
    else // true run - use MPI cached results
    {
        // Gauss integration
        MemTri3 &mem = getMem();
        for (size_t k = 0; k < gauss.getN(); ++k)
        {
            // Jacobian inverse, shape functions and function evaluation
            Eigen::Matrix2d const &J = mem.getJinv(k);
            Eigen::MatrixXd const &dff = cache.getDsf(k);
            Eigen::MatrixXd fk;
            fct.eval(*this, u, k, fk, fake);

            // Elementary stiffness matrix
            K += (fk * J * dff).transpose() * J * dff * gauss.getW(k) * mem.getDetJ(k);
        }
    }
    return K;
}

/**
 * @brief Compute mechanical stiffness matrix
 *
 * K_ij(6,6) = int eps_k H_ijkl eps_l dV = B(6,3)H(3,3)B(3,6)
 */
Eigen::MatrixXd Tri3::buildK_mechanical(Eigen::MatrixXd const &H)
{
    CacheTri3 &cache = getCache(order);
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd K = Eigen::MatrixXd::Zero(6, 6);
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian inverse
        Eigen::Matrix2d const &invJ = mem.getJinv(k);
        // Fill B matrix (shape functions)
        Eigen::MatrixXd const &dff = cache.getDsf(k);
        Eigen::MatrixXd B = Eigen::MatrixXd::Zero(3, 6);
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            // Compute [Ni,x Ni,y]
            Eigen::Vector2d dN = invJ * dff.col(i);
            B(0, 2 * i) = dN(0);
            B(1, 2 * i + 1) = dN(1);
            B(2, 2 * i + 1) = dN(0);
            B(2, 2 * i) = dN(1);
        }

        // Compute stiffness matrix
        K += B.transpose() * H * B * gauss.getW(k) * mem.getDetJ(k);
    }
    return K;
}

/**
 * @brief Compute flux vector for a scalar field
 *
 * s_i(3) = int f*N_i dV
 */
Eigen::VectorXd Tri3::builds(std::vector<double> const &u, Fct0 const &f)
{
    CacheTri3 &cache = getCache(order);
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd s = Eigen::Vector3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
        s += cache.getSf(k) * f.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    return s;
}

/**
 * @brief Compute flux vector for scalar field
 *
 * s_i(3) = int f_j*n_j * N_i dV
 */
Eigen::VectorXd Tri3::builds(std::vector<double> const &u, Fct1 const &f)
{
    CacheTri3 &cache = getCache(order);
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd s = Eigen::Vector3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
        s += cache.getSf(k) * (f.eval(*this, u, k).dot(normal())) * gauss.getW(k) * mem.getDetJ(k);
    return s;
}

/**
 * @brief Compute mass matrix
 *
 * S_ij(3,3) = int N_i*N_j dV
 */
Eigen::MatrixXd Tri3::buildS()
{
    CacheTri3 &cache = getCache(order);
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd S = Eigen::Matrix3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Shape functions
        Eigen::VectorXd const &ff = cache.getSf(k);

        // Mass matrix
        S += ff * ff.transpose() * gauss.getW(k) * mem.getDetJ(k);
    }
    return S;
}

/**
 * @brief Compute gradient of a scalar field at Gauss point k
 */
Eigen::VectorXd Tri3::computeGrad(std::vector<double> const &u, size_t k) const
{
    // Solution vector
    Eigen::Vector3d ue;
    for (size_t i = 0; i < nodes.size(); ++i)
        ue(i) = u[nodes[i]->row];
    // Gradient in global axis: inv(J)_ij * d_jN_i * ue_i
    return getMem().getJinv(k) * getCache(order).getDsf(k) * ue;
}

/**
 * @brief Compute square of gradient of a scalar field at Gauss point k
 */
double Tri3::computeGrad2(std::vector<double> const &u, size_t k) const
{
    return computeGrad(u, k).squaredNorm();
}

/**
 * @brief Compute flux
 *
 * qV_i(3) = int [f] gradu N_i dV
 */
Eigen::VectorXd Tri3::computeqint(std::vector<double> const &u, Fct2 const &fct)
{
    CacheTri3 &cache = getCache(order);
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd qV = Eigen::Vector3d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Shape functions, gradient and flux evaluation
        Eigen::MatrixXd const &dff = cache.getDsf(k);
        Eigen::MatrixXd fk(2, 2);
        fct.eval(*this, u, k, fk, false);

        // Elementary flux vector
        qV += (fk * computeGrad(u, k)).transpose() * mem.getJinv(k) * dff * gauss.getW(k) * mem.getDetJ(k);
    }
    return qV;
}

/**
 * @brief Compute flux
 *
 * qV_i(3) = int -[f] gradu dV
 */
Eigen::VectorXd Tri3::computeqV(std::vector<double> const &u, Fct2 const &fct)
{
    CacheTri3 &cache = getCache(order);
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd qV = Eigen::Vector2d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Gradient and matrix function evaluation
        Eigen::VectorXd gradk = computeGrad(u, k);
        Eigen::MatrixXd fk(2, 2);
        fct.eval(*this, u, k, fk, false);

        qV -= fk * gradk * gauss.getW(k) * mem.getDetJ(k);
    }
    return qV;
}

/**
 * @brief Compute intergal of scalar function
 *
 * I = int f dV
 */
double Tri3::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTri3 &cache = getCache(order);
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    double V = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        V += fct.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    return V;
}

/**
 * @brief Compute intergal of matrix function
 *
 * I = int [f] dV
 */
Eigen::MatrixXd Tri3::computeV(std::vector<double> const &u, Fct2 const &fct)
{
    CacheTri3 &cache = getCache(order);
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    // Gauss integration
    Eigen::MatrixXd out = Eigen::Matrix2d::Zero(); /* @todo out should be resized to match fk size */
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Function evaluation
        Eigen::MatrixXd fk;
        fct.eval(*this, u, k, fk, false);

        out += fk * gauss.getW(k) * mem.getDetJ(k);
    }
    return out;
}

/**
 * @brief Return the element surface
 */
double Tri3::computeV()
{
    return getMem().getVol();
}
